/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)user.c 1.11      04/02/21 Chris G. Davis"

/** \file src/fsm/user.c
 *  \brief FSM User code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "lous/fsm.h"
#include "internal/fsm.h"

fsm_user_t* fsm_user_alloc(fsm_t* fsm, char* tag) {
  fsm_user_t* retval = NULL;
  fsm_trigger_t* walker;
  fsm_event_param_t* param;
  iterator_t* iter;

  LDFunc();
  LDParam(fsm, == NULL, "%p", NULL); 

  if( fsm->lock == false ) {
    LError("fsm not locked");
    errno = EPERM;
    return NULL;
  }

  Lmalloc( retval, sizeof( fsm_user_t ), NULL );

  retval->trigger_state = hash_alloc(8, 0,
				   HASH_MULTIPLEX | HASH_SPLAY | LIST_KEEPFIFO,
				   NULL, NULL, NULL );

  if( retval->trigger_state == NULL ) {
    LError("unable to malloc trigger states");
    free( retval );
    return NULL;
  }
  
  retval->fsm = fsm;

  Lstrdup(retval->tag, tag, NULL, 
	  hash_destroy( retval->trigger_state ); free( retval ); );

  hash_insert( fsm->users, tag, retval);

  fsm->num_users++;

  /* setup trigger list */
  iter = iterator_alloc( fsm->triggers );
  while( (walker = iterator_getnext(iter)) != NULL ) {

    param = fsm_trigger_build_param( walker );
    if( param == NULL ) {
      LError("unable to duplicate trigger list");
      fsm_user_destroy(retval);
      errno = ENOMEM;
      return NULL;
    }

    hash_insert(retval->trigger_state, param->trigger_tag, param);
  }
  iterator_destroy( iter );

  retval->curr_state = fsm->initial;
  retval->prev_state = NULL;
  retval->been_run = false;

  return retval;
}

void fsm_user_destroy(fsm_user_t* user) {
  fsm_event_param_t* param;
  iterator_t* iter;

  LDFunc();
  LDParam(user, == NULL, "%p");

  hash_remove(user->fsm->users, user->tag);
  user->fsm->num_users--;

  if( user->tag ) {
    free( user->tag );
  }

  if( user->trigger_state ) {

    iter = iterator_alloc( user->trigger_state );
    while( (  param = iterator_getnext( iter )) != NULL ) {
      hash_remove( user->trigger_state, param->trigger_tag );

      if( param->trigger_tag ) {
	free( param->trigger_tag );
      }

      free( param ); 
    }

    hash_destroy( user->trigger_state );

    iterator_destroy( iter );
  }

  free( user );
}

void fsm_user_set_trigger(fsm_user_t* user, char* tag, s32_t val) {
  fsm_event_param_t* trigger;

  LDFunc();
  LDParam(user, == NULL, "%p");

  trigger = hash_lookup(user->trigger_state, tag);

  if( trigger == NULL ) {
    LError("Trigger %s not found", tag);
    errno = ENOENT;
  } else {
    trigger->val = val;
  }
  
}

s32_t fsm_user_get_trigger(fsm_user_t* user, char* tag) {
  fsm_event_param_t* trigger;

  LDFunc();
  LDParam(user, == NULL, "%p", -1);

  trigger = hash_lookup(user->trigger_state, tag);

  if( trigger == NULL ) {
    LError("Trigger %s not found", tag);
    errno = ENOENT;
    return 0;
  } else {
    return trigger->val;
  }
  
}

void fsm_user_inc_trigger(fsm_user_t* user, char* tag, s32_t val) {
  fsm_event_param_t* trigger;

  LDFunc();
  LDParam(user, == NULL, "%p");

  trigger = hash_lookup(user->trigger_state, tag);

  if( trigger == NULL ) {
    LError("Trigger %s not found", tag);
    errno = ENOENT;
  } else {
    trigger->val+= val;
  }
  
}

void fsm_user_run(fsm_t* fsm, fsm_user_t* user) {
  fsm_state_t* state;
  fsm_event_t* event;
  struct timeval currtime;
  iterator_t* iter;

  LDFunc();
  LDParam(fsm, == NULL, "%p");
  LDParam(user, == NULL, "%p");

  state = user->curr_state;

  /* check for timeout */
  if( state->flags & FSM_STATE_TIMEOUT ) {
    gettimeofday(&currtime, NULL);

    if( ( currtime.tv_sec - user->last_state_change.tv_sec ) > state->val ) {
      /* timed out */
      fsm_user_set_trigger(user, "timeout", 1);
    }
  }

  /* run the callback on the current state */
  if( state->callback && ( user->been_run == false )) {
    state->callback(fsm, user->tag);
    user->been_run = true;
  }
  /*
   * Now, update the user
   *
   * How, find the first event that matches
   */
  iter = iterator_alloc( state->events );
  while( ( event = iterator_getnext( iter )) != NULL ) {
    if( fsm_event_matches( event, user->trigger_state) == 0 ) {
      /* this matches */
      user->prev_state = user->curr_state;
      user->curr_state = event->target;
      gettimeofday( &user->last_state_change, NULL );

      if ( fsm->callback ) {
	fsm->callback(fsm, user->tag, user->curr_state->tag, user->prev_state->tag);
      }
      user->been_run = false;

      break;
    }
  }
  iterator_destroy( iter );

}

void fsm_user_display(fsm_user_t* user) {
  fsm_event_param_t* param;
  iterator_t* iter;

  printf("User: %s\n", user->tag);
  printf("  Curr: %-30s Prev: %-30s\n", 
	 (user->curr_state)?user->curr_state->tag:"<NULL>", 
	 (user->prev_state)?user->prev_state->tag:"<NULL>" );

  iter = iterator_alloc(user->trigger_state);
  while( ( param = iterator_getnext( iter )) != NULL ) {
    printf("    %s = %d\n", param->trigger_tag, param->val);
  }

  iterator_destroy( iter );
}

void fsm_user_force_state(fsm_user_t* user, char* state) {
  LDFunc();
  LDParam(user, == NULL, "%p");
  LDParam(state, == NULL, "%s");

  user->prev_state = user->curr_state;
  if( (user->curr_state = fsm_state_find(user->fsm, state)) == NULL ) {
    LError("State %s not found", state);
  }

  gettimeofday( &user->last_state_change, NULL );
  
  if ( user->fsm->callback ) {
    user->fsm->callback(user->fsm, user->tag, user->curr_state->tag, user->prev_state->tag);
  }
  user->been_run = false;
}
