/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_INCLUDE_HASH_H
#define LOUS_SRC_INCLUDE_HASH_H

#ident "@(#)hash.h 1.9      04/02/16 Chris G. Davis"

/** \file src/include/hash.h
 *  \brief Hash tables exported definitions
 *  \author Chris G. Davis
 *
 */

/*
 * Required LOUS functionality
 */
#include "lous/iterator.h"

struct LOUS_Hash_Table;
typedef struct LOUS_Hash_Table hash_t;

struct LOUS_Hash_Entry;
typedef struct LOUS_Hash_Entry hash_ent_t;

struct LOUS_Hash_Operations;
typedef struct LOUS_Hash_Operations hash_ops_t;

#define HASH_MULTIPLEX     0x01  /**< map multiple keys to one hash bucket */

/* 
 * When HASH_MULTIPLEX is specified
 *  key in hash_ent_t is invalid, and data in hash_ent_t points to
 *  an additional data structure which actually contains the key,data pair
 *
 */
#define HASH_LINKEDLIST    0x10 /**< Each bucket is a simple linked list */
#define HASH_SPLAY         0x20 /**< Each bucket is a splay tree */

/** allocate a new hash
 *
 * This allocates a new hash, and initializes it according to 
 * the specified flags
 *
 * \param buckets approximate number of buckets.
 * \param keylen length of the keys, or 0 if the keys are strings
 * \param flags hash flags
 * \param hash_func function to compute the bucket for the given key
 * \param comp_func function to compare 2 keys
 * \param free_func function to use when deleteing data items, NULL if caller will
 *
 * \return hash is successful, NULL if failure, with errno set
 */
hash_t* hash_alloc(s32_t buckets, s32_t keylen, s32_t flags,
		   s32_t (*hash_func)(void*, s32_t, s32_t),
		   comp_t (*comp_func)(void*, void*, s32_t),
		   void (*free_func)(void*));

/** destroy the hash
 *
 * Destory the hash, calling destory. Any data still in the hash 
 * is free'd if free_func is non-null
 *
 * \param hash the hash to destroy
 *
 * \return 0 if successful, -1 if failure, with errno set.
 */
s32_t hash_destroy(hash_t* hash);

/** insert an entry into the hash.
 *
 * The hash must be predefined. Key is copied and stored, but
 * data is used uncopied and must be allocated and freed by the
 * calling program.
 *
 * \param hash the hash tree to insert into
 * \param key the key to use, a string or hash->key_size bytes
 * \param data the data to insert into the hash
 *
 * \return 0 if successful, -1 if failure, with errno set
 */
s32_t hash_insert(hash_t* hash, void* key, void* data);

/** remove an entry from the hash.
 *
 * The hash must be predefined. 
 *
 * \param hash the hash tree to remove from
 * \param key the key to use, a string or hash->key_size bytes
 *
 * \return data successful, NULL if failure, with errno set
 */
void* hash_remove(hash_t* hash, void* key);

/** remove and delete an entry from the hash.
 *
 * The hash must be predefined, data is deleted is free_func
 * is non-null
 *
 * \param hash the hash tree to remove from
 * \param key the key to use, a string or hash->key_size bytes
 *
 * \return 0 if successful, -1 if failure, with errno set
 */
s32_t hash_delete(hash_t* hash, void* key);

/** replace an entry in the hash.
 *
 * The hash must be predefined. Key is unchanged, but
 * data is used uncopied and must be allocated by the
 * calling program.
 *
 * \param hash the hash tree to replace in
 * \param key the key to use, a string or hash->key_size bytes
 * \param data the new data to replace
 *
 *\return old data if successful, NULL if failure, with errno set
 */
void* hash_replace(hash_t* hash, void* key, void* data);

/** lookup an entry in the hash.
 *
 * The hash must be predefined.
 *
 * \param hash the hash tree to replace in
 * \param key the key to use, a string or hash->key_size bytes
 *
 *\return data if successful, NULL if failure, with errno set
 */
void* hash_lookup(hash_t* hash, void* key);

/** get all keys from the hash.
 *
 * The hash must be predefined.
 * 
 * \param hash the hash tree to get list of keys from
 *
 * \return list of keys, or NULL if failure
 */
void** hash_getkeys(hash_t* hash);

/** Get the current number of entries in the hash
 *
 */
s32_t hash_numentries(hash_t* hash);

/** Get the number of buckets in the hash
 *
 */
s32_t hash_get_buckets(hash_t* hash);

/** Get the data in the specified bucket
 *
 */
void* hash_get_bucket(hash_t* hash, s32_t bucket);


/*
 * util functions in util.c
 */
comp_t default_comp(void* keya, void* keyb, s32_t keylen);
s32_t default_hash(void* key, s32_t keylen, s32_t buckets);

#endif /* LOUS_SRC_INCLUDE_HASH_H */
