/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_INCLUDE_ITERATOR_H
#define LOUS_SRC_INCLUDE_ITERATOR_H

#ident "@(#)iterator.h 1.2      04/02/21 Chris G. Davis"

/** \file src/include/iterator.h
 *  \brief iterator functionality
 *  \author Chris G. Davis
 *
*/

/*
 * Required LOUS functionality
 */
#include "lous/lous.h"

/**
 *
 * lists must be of the format
 * struct some_list {
 *     iterator_list_t iterate;
 *    comp_t (*comp_func);
 *     ... more data items ....
 * }
 *
 * Note that the data items must be of the format:
 *
 * struct some_data {
 *    iterator_entry_t iterate;
 *    ... more data items ...
 * }
 *
 * for sorted or reverse sorted data, we need
 * struct some_data {
 *    iterator_entry_t iterate;
 *    char key[];
 *    ... more data items ...
 * }
 * or
 * struct some_data {
 *    iterator_entry_t iterate;
 *    char* key;
 *    ... more data items ...
 * }
 *
 */

/* structure def */
struct LOUS_Iterator;
typedef struct LOUS_Iterator iterator_t; /**< iterator structure */

struct LOUS_Iterator_Entry;
typedef struct LOUS_Iterator_Entry iterator_entry_t; /**< iterator entry structure */

struct LOUS_Iterator_Entry {
  iterator_entry_t* next;
  iterator_entry_t* prev;

};

struct LOUS_Iterator_List;
typedef struct LOUS_Iterator_List iterator_list_t; /**< iterator list structure */

struct LOUS_Iterator_List {
  iterator_entry_t* first;
  iterator_entry_t* end;

};


/* defines */
#define LIST_KEEPFIFO      0x1000 /**< keep a FIFO list of all entries */
#define LIST_KEEPLIFO      0x2000 /**< keep a LIFO list of all entries */
#define LIST_KEEPSORTED    0x4000 /**< keep a sorted list of all entries */
#define LIST_KEEPREVERSE   0x8000 /**< keep a reverses sorted list of all entries */

#define LIST_ITERATOR      0xf000 /**< all list types */
#define LIST_HASKEY        0xb000 /**< has a key entry */

#define LIST_KEYPTR        0x0100 /**< key is a ptr, not part of data */

/* functions */
/** Remove an entry from the iterator list
 * 
 * @param flags
 * @param list
 * @param ent
 */
void iterator_do_remove(s32_t flags, iterator_list_t* list, iterator_entry_t* ent);

/** Insert an entry into the iterator list
 * 
 * @param flags
 * @param list
 * @param ent
 */
void iterator_do_insert(s32_t flags, iterator_list_t* list, iterator_entry_t* ent);

/** Allocate a new iterator
 * 
 * @param list
 */
iterator_t* iterator_do_alloc(iterator_list_t* list);

/** Destroy the iterator
 * 
 * @param iter
 */
void iterator_destroy(iterator_t* iter);

/** Get the first entry in the iterator list
 * 
 * @param iter
 */
void* iterator_getfirst(iterator_t* iter);

/** Get the last entry in the iterator list
 * 
 * @param iter
 */
void* iterator_getlast(iterator_t* iter);

/** Get the next entry in the iterator list
 * 
 * @param iter
 */
void* iterator_getnext(iterator_t* iter);

/** Get the previous entry in the iterator list
 * 
 * @param iter
 */
void* iterator_getprev(iterator_t* iter);

/* macros */
/** Allocate an iterator
 * 
 */
#define iterator_alloc(list) \
 iterator_do_alloc((iterator_list_t*)list)

/** Destroy and iterator
 * 
 */
#define iterator_remove(which, wdata) \
 iterator_do_remove(which->flags, (iterator_list_t*)which, (iterator_entry_t*)wdata)

/** Insert an entry to the iterator 
 * 
 */
#define iterator_insert(which, wdata) \
 iterator_do_insert(which->flags, (iterator_list_t*)which, (iterator_entry_t*)wdata)

/** initialize the iterator list structure 
 * 
 */
#define iterator_list_initialize(which) \
(iterator_list_t*)which->iterate.first = NULL; \
(iterator_list_t*)which->iterate.end = NULL; 


#endif /* LOUS_SRC_INCLUDE_ITERATOR_H */
