/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_INCLUDE_LOGGING_H
#define LOUS_SRC_INCLUDE_LOGGING_H

#ident "@(#)logging.h 1.12      04/02/23 Chris G. Davis"

/** \file src/include/logging.h
 *  \brief Logging and debugging definitions
 *  \author Chris G. Davis
 *
 * \todo test macros
 */

#include <errno.h>

#define LOGGING_MAX_GROUPS 32 /**< Maximum 32 groups (32bit integer) */
#define LOGGING_MAX_LEVELS 7 /**< maximum levels, just because */

#define LOG_STDOUT     0x0001 /**< log to stdout */
#define LOG_STDERR     0x0002 /**< log to stderr */
#define LOG_FLUSH      0x0004 /**< flush after each log message */
#define LOG_NODATE     0x0010 /**< don't display the data */
#define LOG_NOTIME     0x0020 /**< don't display the time */
#define LOG_FILE       0x0100 /**< log to a file (pass in the path) */
#define LOG_SOCKET     0x0200 /**< log to a file descriptor */

struct LOUS_Logging_Group;
typedef struct LOUS_Logging_Group loggroup_t; /**< logging group structure */

/** Initialize a new logging group
 * 
 * \return new loggroup or NULL on failure.
 */
loggroup_t* loggroup_init(bool_t ft);

/** Destroy a log group
 * 
 * @param group
 */
s32_t loggroup_destroy(loggroup_t* group);

/** Register a new class in the group
 * 
 * @param group
 * @param name
 *
 * @return the class number
 */
s32_t loggroup_reg_class(loggroup_t* group, char* name);

/** Register a new level in the group
 * 
 * @param group
 * @param name
 *
 * @return the level number
 */
s32_t loggroup_reg_level(loggroup_t* group, char* name);

/** Remove a class from the group
 * 
 * @param group
 * @param index
 */
s32_t loggroup_dereg_class(loggroup_t* group, s32_t index);

/** Remove a class by name from the group
 * 
 * @param group
 * @param name
 */
s32_t loggroup_dereg_class_name(loggroup_t* group, char* name);

/** Remove a level from the group
 * 
 * @param group
 * @param index
 */
s32_t loggroup_dereg_level(loggroup_t* group, s32_t index);

/** Remove a level by name from the group
 * 
 * @param group
 * @param name
 */
s32_t loggroup_dereg_level_name(loggroup_t* group, char* name);

/** Update the current time timeval using within the logging
 * 
 * @param group
 */
void loggroup_update_currtime(loggroup_t* group);

typedef s32_t (*logcallback_t)(loggroup_t* group, s32_t class, s32_t level,
				char* msg, ...);

struct LOUS_Logging_User;
typedef struct LOUS_Logging_User loguser_t; /**< log user structure */

/** Initialize a new log user
 * 
 * @param group
 * @param flags
 * @param lout
 *
 * @return the log user structure
 */
loguser_t* loguser_init(loggroup_t* group, s32_t flags, void* lout);

/** Destroy the loguser
 * 
 * @param user
 */
s32_t loguser_destory(loguser_t* user);

/** Reopen any open log files
 *
 * @param user
 * @param path
 */
void loguser_reopen(loguser_t* user, char* path);

/** Modify the user's class state
 * 
 * @param user
 * @param class
 * @param state
 */
s32_t loguser_mask(loguser_t* user, s32_t class, state_t state);

/** Modify the user's class log level
 * 
 * @param user
 * @param class
 * @param state
 * @param val
 */
s32_t loguser_level(loguser_t* user, s32_t class, state_t state, s32_t val);

/** Set the users class callback 
 * 
 * @param user
 * @param class
 * @param callback
 */
s32_t loguser_set_callback(loguser_t* user, s32_t class, 
			    logcallback_t* callback);

/** Remove the users class callback
 * 
 * @param user
 * @param class
 */
s32_t loguser_clear_callback(loguser_t* user, s32_t class);

/** adjust the users flags
 * 
 * @param user
 * @param flag
 * @param state
 */
bool_t loguser_flag(loguser_t* user, s32_t flag, state_t state);


/** log a message
 * 
 * @param grp
 * @param class
 * @param level
 * @param msg
 */
void logging_log(loggroup_t* grp, s32_t class, s32_t level, char* msg, ...);

/** send a single debug message
 * 
 */
#define Debug(grp, class, level, stuff, ...) \
logging_log(grp, class, level, #stuff, ##__VA_ARGS__)

/** send an Error message, including location and error info
 * 
 */
#define Error(grp, msg, ...) \
logging_log(grp, -1, 0, "[%s %d %s]", __FILE__, __LINE__, __FUNCTION__); \
logging_log(grp, -1, 0, " (%d) %s", errno, strerror(errno)); \
logging_log(grp, -1, 0, #msg, ##__VA_ARGS__)



#ifdef LOUS_DEBUG
extern s32_t LD_func; /**< LOUS function debugging */
extern s32_t LD_debug; /**< LOUS other debugging */
extern s32_t LL_trace; /**< LOUS trace level logging */
extern s32_t LL_verbose; /**< LOUS verbose level logging */
extern s32_t LL_veryverbose; /**< LOUS very verbose level logging */
extern loggroup_t* lous_loggroup; /**< LOUS logging group */
extern loguser_t* lous_loguser; /**< LOUS logging user */

/** LOUS Debugging message
 * 
 */
#define LDebug(class, level, stuff, ...) \
  logging_log(lous_loggroup, class, level, #stuff, ##__VA_ARGS__)

/** LOUS Error message
 * 
 */
#define LError(msg, ...) \
logging_log(lous_loggroup, -1, 0, "[%s %d %s]", __FILE__, __LINE__, __FUNCTION__); \
logging_log(lous_loggroup, -1, 0, " (%d) %s", errno, strerror(errno)); \
logging_log(lous_loggroup, -1, 0, #msg, ##__VA_ARGS__)

/** LOUS Function debugging
 * 
 */
#define LDFunc()\
logging_log(lous_loggroup, LD_func, LL_verbose, "In function %s [%s:%d]", \
__FUNCTION__, __FILE__, __LINE__); \
errno=0

/** LOUS function Parameter debugging
 * 
 */
#define LDParam(param, cond, how, ...) \
if( param  cond ) {\
  errno = EINVAL; \
  logging_log(lous_loggroup, -1, 0, "Param Failed check: " #param #cond ); \
  logging_log(lous_loggroup, -1, 0, "Param had value: "  how, param); \
  return __VA_ARGS__; \
} else { \
logging_log(lous_loggroup, LD_func, LL_veryverbose, #param " = " how, param); \
} \
while( 0 ) {}

/** Initialize the LOUS debugging 
 * 
 */
#define LOUS_Debug_Init(flags, nflgs) \
lous_loggroup = loggroup_init(true); \
LD_func = loggroup_reg_class(lous_loggroup, "func  "); \
LD_debug = loggroup_reg_class(lous_loggroup, "debug  "); \
LL_trace = loggroup_reg_level(lous_loggroup, "trace   "); \
LL_verbose = loggroup_reg_level(lous_loggroup, "verbose "); \
LL_veryverbose = loggroup_reg_level(lous_loggroup, "vverbose"); \
lous_loguser = loguser_init(lous_loggroup, 0, NULL); \
loguser_mask(lous_loguser, LD_func, set); \
loguser_mask(lous_loguser, LD_debug, set); \
loguser_flag(lous_loguser, flags, set); \
loguser_flag(lous_loguser, nflgs, clear)

/** LOUS set debugging info
 * 
 */
#define LOUS_SetDebug(cls, lvl) \
loguser_level(lous_loguser, cls, set, lvl)

#else
#define LD_func 0
#define LD_debug 0
#define LL_trace 0
#define LL_verbose 0
#define LL_veryverbose 0

#define LDebug(...)
#define LDFunc(...) errno = 0
#define LError(...)

#define LDParam(param, cond, how, ...) \
if( param cond ) { \
  errno = EINVAL; \
  return __VA_ARGS__; \
} \
while( 0 ) {}

#define LOUS_Debug_Init(...)

#define LOUS_SetDebug(...)

#endif /* LOUS_DEBUG */

#endif /* LOUS_SRC_INCLUDE_LOGGING_H */
