/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_INCLUDE_NETEVENT_H
#define LOUS_SRC_INCLUDE_NETEVENT_H

#ident "@(#)netevent.h 1.7      04/02/21 Chris G. Davis"

/** \file src/include/netevent.h
 *  \brief Network event watching and callbacks
 *  \author Chris G. Davis
 *
*/

/*
 * Required LOUS functionality
 */
#include "lous/hash.h"

#define NEVNT_OWNFD         0x0001 /**< close fd's on remove or destroy */

#define NEVNT_NONBLOCK    0x0010 /**< force fd's to be non-blocking */
#define NEVNT_WATCHREAD   0x0100 /**< watch for read ready */
#define NEVNT_WATCHWRITE  0x0200 /**< watch for write ready */
#define NEVNT_WATCHEXCEPT 0x0400 /**< watch for exceptions */

#define NEVNT_TIMEOUT     0x1000 /**< we're supplying a timeout */
#define NEVNT_IDLETIMEOUT 0x2000 /**< timeout only if connection idle */

/*
 * Values sent to callback
 */
#define NEVNT_GOTREAD       0x0001 /**< read events are waiting */
#define NEVNT_GOTWRITE      0x0002 /**< write events are waiting */
#define NEVNT_GOTEXCEPT     0x0004 /**< exception events are waiting */
#define NEVNT_CLOSED        0x0008 /**< the netevent is closed */

struct LOUS_Netevent;
typedef struct LOUS_Netevent netevent_t; /**< netevent structure */

struct LOUS_Netevent_Class;
typedef struct LOUS_Netevent_Class netevent_class_t; /**< netevent class */

struct LOUS_Netevent_User;
typedef struct LOUS_Netevent_User netevent_user_t; /**< netevent user */

/** Allocate a netevent structure
 * 
 * @param flags
 * @param timeout
 * @param sec
 * @param usec
 *
 * @return the netevent structure
 */
netevent_t* netevent_alloc(s32_t flags, s32_t timeout, s32_t sec, s32_t usec);

/** Destroy the netevent 
 * 
 * @param event
 */
s32_t netevent_destroy(netevent_t* event);

/** Run through the classes and users, and check for updates
 * 
 * @param event
 */
s32_t netevent_run(netevent_t* event);

/** Add a new class to the event
 * 
 * @param event
 * @param tag
 * @param flags
 * @param timeout
 * @param callback
 *
 * @return the netevent class structure
 */
netevent_class_t* netevent_addclass(netevent_t* event, char* tag, s32_t flags,
				    s32_t timeout,
				    s32_t (*callback)(s32_t fd, s32_t flag, char* tag));

/** Remove and destroy the class from the event
 * 
 * @param event
 * @param class
 */
s32_t netevent_remclass(netevent_t* event, netevent_class_t* class);

/** Remove and destroy the named class from the event
 * 
 * @param event
 * @param tag
 */
s32_t netevent_remclass_name(netevent_t* event, char* tag);

/** Add a new user to the class 
 * 
 * @param class
 * @param tag
 * @param timeout
 * @param flags
 * @param fd
 *
 * @return the netevent user structure
 */
netevent_user_t* netevent_adduser(netevent_class_t* class, char* tag,
				  s32_t timeout,
				  s32_t flags, s32_t fd);

/** Remove the user from the class
 * 
 * @param class
 * @param user
 */
s32_t netevent_remuser(netevent_class_t* class, netevent_user_t* user);

/** Remove the named user from the class
 * 
 * @param class
 * @param tag
 */
s32_t netevent_remuser_name(netevent_class_t* class, char* tag);

#endif /* LOUS_SRC_INCLUDE_NETEVENT_H */
