/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)logging.c 1.13      04/02/23 Chris G. Davis"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "internal/logging.h"

/** \file src/logging/logging.c
 *  \brief General logging and debugging code
 *  \author Chris G. Davis
 *
 * \todo Finish callbacks
 */

s32_t loggroup_find_class_name(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t i;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    for(i=0; i< LOGGING_MAX_GROUPS; i++ ) {
      if( group->classes[i] != NULL ) {
	if( !strcmp(name, group->classes[i])) {
	  retval = i;
	  break;
	}
      }
    }
  }

  return retval;
}

s32_t loggroup_find_level_name(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t i;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    for(i=0; i< LOGGING_MAX_LEVELS; i++ ) {
      if( group->levels[i] != NULL ) {
	if( !strcmp(name, group->levels[i])) {
	  retval = i;
	  break;
	}
      }
    }
  }

  return retval;
}

s32_t loggroup_find_empty_class(loggroup_t* group) {
  s32_t retval = -1;
  s32_t i;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    for(i=0; i< LOGGING_MAX_GROUPS; i++ ) {
      if( group->classes[i] == NULL ) {
	retval = i;
	break;
      }
    }
  }
  return retval;
}

s32_t loggroup_find_empty_level(loggroup_t* group) {
  s32_t retval = -1;
  s32_t i;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    for(i=0; i< LOGGING_MAX_LEVELS; i++ ) {
      if( group->levels[i] == NULL ) {
	retval = i;
	break;
      }
    }
  }
  
  return retval;
}

loggroup_t* loggroup_init(bool_t ft) {
  loggroup_t* newgroup = NULL;

  errno = 0;

  if( (newgroup = malloc(sizeof( loggroup_t ))) != NULL ) {
    memset(newgroup, 0, sizeof( loggroup_t ));
    newgroup->num_classes = 0;
    newgroup->num_levels = 0;
    newgroup->num_users = 0;
    newgroup->users = NULL;
    newgroup->fasttime = ft;
  }

  return newgroup;
}

s32_t loggroup_destroy(loggroup_t* group) {
  s32_t retval = -1;

  errno = 0;

  if ( group == NULL ) {
    errno = ENOENT;
  } else {
    if( group->users > 0 ) {
      errno = EBUSY;
    } else {
      /* nothing really to do */
      retval = 0;
    }
  }

  return retval;
}

s32_t loggroup_reg_class(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t nextent;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( name == NULL ) {
      errno = EINVAL;
    } else {
      if ( loggroup_find_class_name(group, name) != -1 ) {
	errno = EEXIST;
      } else {
	nextent = loggroup_find_empty_class(group);
	if ( nextent == -1 ) {
	  errno = EACCES;
	} else { 
	  group->classes[nextent] = name;
	  group->num_classes++;
	  retval = nextent;
	}
      }
    }
  }

  return retval;
}

s32_t loggroup_reg_level(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t nextent;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( name == NULL ) {
      errno = EINVAL;
    } else {
      if ( loggroup_find_level_name(group, name) != -1 ) {
	errno = EEXIST;
      } else {
	nextent = loggroup_find_empty_level(group);
	if ( nextent == -1 ) {
	  errno = EACCES;
	} else { 
	  group->levels[nextent] = name;
	  group->num_levels++;
	  retval = nextent;
	}
      }
    }
  }

  return retval;
}

s32_t loggroup_dereg_class(loggroup_t* group, s32_t index) {
  s32_t retval = -1;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( group->classes[index] == NULL ) {
	errno = ENOENT;
    } else {
      group->classes[index] = NULL;
      group->num_levels--;
      retval = 0;
    }
  }

  return retval;
}

s32_t loggroup_dereg_class_name(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t ent;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( (ent = loggroup_find_class_name(group, name)) == -1 ) {
      errno = ENOENT;
    } else {
      group->classes[ent] = NULL;
      group->num_levels--;
      retval = 0;
    }
  }

  return retval;
}

s32_t loggroup_dereg_level(loggroup_t* group, s32_t index) {
  s32_t retval = -1;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( group->levels[index] == NULL ) {
	errno = ENOENT;
    } else {
      group->levels[index] = NULL;
      group->num_levels--;
      retval = 0;
    }
  }

  return retval;
}

s32_t loggroup_dereg_level_name(loggroup_t* group, char* name) {
  s32_t retval = -1;
  s32_t ent;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if ( (ent = loggroup_find_level_name(group, name)) == -1 ) {
      errno = ENOENT;
    } else {
      group->classes[ent] = NULL;
      group->num_levels--;
      retval = 0;
    }
  }

  return retval;
}

void loggroup_update_currtime(loggroup_t* group) {
  if( group == NULL ) {
    return;
  }

  gettimeofday(&group->currtime, NULL);
}

loguser_t* loguser_init(loggroup_t* group, s32_t flags, void* lout) {
  loguser_t* newuser = NULL;

  errno = 0;

  if( group == NULL ) {
    errno = ENOENT;
  } else {
    if( (newuser = malloc( sizeof( loguser_t ))) != NULL ) {
      memset(newuser, 0, sizeof( loguser_t ));
      newuser->group = group;

      newuser->classmask = 0;
      newuser->user_callback = NULL;

      if( flags & LOG_STDOUT ) {
	newuser->log_stdout = true;
      } else {
	newuser->log_stdout = false;
      }

      if( flags & LOG_STDERR ) {
	newuser->log_stderr = true;
      } else {
	newuser->log_stderr = false;
      }

      if( flags & LOG_FLUSH ) {
	newuser->log_flush = true;
      } else {
	newuser->log_flush = false;
      }

      if( flags & LOG_FILE ) {
	if( ( newuser->logout = fopen(lout, "a+")) == NULL ) {
	  fprintf(stderr, "unable to open %s for logging: %s\n", (char*)lout, strerror(errno));
	  free( newuser );
	  return NULL;
	}
      } else {
	newuser->logout = lout;
      }
      newuser->group->num_users++;
      newuser->next = newuser->group->users;
      newuser->group->users = newuser;
      newuser->flags = flags;
    }
  }

  return newuser;
}

s32_t loguser_destory(loguser_t* user) {
  s32_t retval = -1;
  loguser_t* curr;

  errno = 0;

  user->group->num_users--;
  if( user->group->users == user ) {
    user->group->users = user->next;
  } else {
    curr = user->group->users;
    while( curr != NULL ) {
      if( curr->next == user ) {
	curr->next = curr->next->next;
	break;
      }
      curr = curr->next;
    }
  }

  retval = 0;

  return retval;
}

void loguser_reopen(loguser_t* user, char* path) {
  if( user == NULL ) {
    return;
  }

  if( path == NULL ) {
    return;
  }

  if( user->logout == NULL ) {
    return;
  }

  if( user->flags & LOG_FILE ) {
    fclose( (FILE*)user->logout );
	
    if( ( user->logout = fopen(path, "a+")) == NULL ) {
      fprintf(stderr, "unable to open %s for logging: %s\n", (char*)path, strerror(errno));
    }
  }
}

s32_t loguser_mask_helper(loguser_t* user, s32_t class, state_t state) {
  s32_t retval = 0;

  errno = 0;

  switch (state) {
  case toggle:
    user->classmask ^= 1 << class;
    break;
  case get:
    if( class == -1 ) {
      retval = user->classmask;
    } else {
      retval = user->classmask & ( 1 << class);
    }
    break;
  case clear:
    user->group->class_users[class]--;
    user->classmask &= ~( 1 << class);
    break;
  case set:
    user->group->class_users[class]++;
    user->classmask |= ( 1 << class);
    break;
  }

  return retval;
}

s32_t loguser_mask(loguser_t* user, s32_t class, state_t state) {
  s32_t retval = -1;
  s32_t count;

  errno = 0;

  if( user == NULL ) {
    errno = ENOENT;
  } else {
    if( class == -1 ) {
      for(count=0; count< LOGGING_MAX_GROUPS; count++) {
	loguser_mask_helper(user, count, state);
      }
      
      retval = 0;

    } else {
      retval = loguser_mask_helper(user, class, state);
    }
  }

  return retval;
}

s32_t loguser_level_helper(loguser_t* user, s32_t class, state_t state, s32_t val) {
  s32_t retval = 0;
  s32_t count;

  switch (state) {
  case toggle:
    retval = -1;
    errno = EINVAL;
    break;
  case get:
    retval = user->class_level[class] - 1;
    break;
  case clear:
    for(count=0; count<val+1; count++) {
      user->group->level_users[count]--;
    }
    user->class_level[class] = 0;
    break;
  case set:
    for(count=0; count<val+1; count++) {
      user->group->level_users[count]++;
    }
    user->class_level[class] = val + 1;
    break;
  }

  return retval;
}

s32_t loguser_level(loguser_t* user, s32_t class, state_t state, s32_t val) {
  s32_t retval = -1;
  s32_t count;

  errno = 0;

  if( user == NULL ) {
    errno = ENOENT;
  } else {
    if( class == -1 ) {
      for(count=0; count< LOGGING_MAX_GROUPS; count++) {
	loguser_level_helper(user, count, state, val);
      }
      
      return 0;

    } else {
      retval = loguser_level_helper(user, class, state, val);
    }
  }

  return retval;
}

s32_t loguser_set_callback(loguser_t* user, s32_t class, 
			   logcallback_t* callback) {
  s32_t retval = -1;

  return retval;
}

s32_t loguser_clear_callback(loguser_t* user, s32_t class) {
  s32_t retval = -1;

  return retval;
}

bool_t loguser_flag_helper(bool_t* flag, state_t state) {
  bool_t retval = false;

  switch (state) {
  case toggle:
    *flag ^= true;
    break;
  case get:
    retval = *flag;
    break;
  case clear:
    *flag = false;
    break;
  case set:
    *flag = true;
    break;
  }

  return retval;
}

bool_t loguser_flag(loguser_t* user, s32_t flag, state_t state) {
  bool_t retval = false;

  errno = 0;

  if( user == NULL ) {
    errno = ENOENT;
  } else {
    if( flag & LOG_STDOUT ) {
      retval = loguser_flag_helper( &user->log_stdout, state);
    }
    if ( flag & LOG_STDERR ) {
      retval = loguser_flag_helper( &user->log_stderr, state);
    }
    if ( flag & LOG_FLUSH ) {
      retval = loguser_flag_helper( &user->log_flush, state);
    }
    if ( flag & LOG_NODATE ) {
      retval = loguser_flag_helper( &user->log_nodate, state);
    }
    if ( flag & LOG_NOTIME ) {
      retval = loguser_flag_helper( &user->log_notime, state);
    }
  }

  return retval;
}

#define LOGGING_MAX_LINE 255
void logging_log_user(loguser_t* user, s32_t class, s32_t level,
		      char* msg, char* indata, char* datebuf, char* timebuf) {
  char buffer[LOGGING_MAX_LINE];

  if( ( class == -1 ) ||
      ( ( user->classmask & ( 1 << class ) ) && 
	( user->class_level[class] > level ))) {
      
    snprintf(buffer, LOGGING_MAX_LINE, 
	     "%s%s%s ] %s", 
	     msg,
	     (user->log_nodate == false)?datebuf:"",
	     (user->log_notime == false)?timebuf:"",
	     indata);
    
    
    if( user->log_stdout == true ) {
      fprintf(stdout, "%s\n", buffer);
      if( user->log_flush == true ) {
	fflush(stdout);
      }
    }
    
    if( user->log_stderr == true ) {
      fprintf(stderr, "%s\n", buffer);
      if( user->log_flush == true ) {
	fflush(stderr);
      }
    }
    
    if( user->logout != NULL ) {
      if( user->flags & LOG_FILE ) {
	fprintf((FILE*)user->logout, "%s\n", buffer);
	if( user->log_flush == true ) {
	  fflush((FILE*)user->logout);
	}
      } else if ( user->flags & LOG_SOCKET ) {
	write((int)user->logout, buffer, strlen( buffer ));
      }
    }
  }
}

void logging_log(loggroup_t* group, s32_t class, s32_t level, char* msg, ...) {
  va_list ap;
  char buffer[LOGGING_MAX_LINE];
  char indata[LOGGING_MAX_LINE];
  char datebuf[13];
  char timebuf[18];
  struct timeval tv;
  struct tm* tmv;
  loguser_t* user;

  if ( group != NULL ) {
    if( group->class_users[class] == 0 ) {
      return;
    }

    if( group->level_users[level] == 0 ) {
      return;
    }
#if 0
    if( group->class_level_users[class][level] == 0 ) {
      return;
    }
#endif

    if( group->fasttime == true ) {
      tmv = localtime(&group->currtime.tv_sec);
      group->currtime.tv_usec++;
    } else {
      gettimeofday(&tv, NULL);
      tmv = localtime(&tv.tv_sec);
    }

    va_start( ap, msg);
    vsnprintf(indata, LOGGING_MAX_LINE, msg, ap);
    va_end(ap);

    snprintf(datebuf, 12, " %02d/%02d/%4d",
	     tmv->tm_mon + 1,
	     tmv->tm_mday,
	     tmv->tm_year + 1900);

    snprintf(timebuf, 17, " %02d:%02d:%02d.%06ld",
	     tmv->tm_hour,
	     tmv->tm_min,
	     tmv->tm_sec,
	     ( group->fasttime == true)?group->currtime.tv_usec:tv.tv_usec);

    snprintf(buffer, LOGGING_MAX_LINE, 
	     "%c[%-s (%1d)", 
	     (class == -1)?'X':' ', 
	     (class != -1)?group->classes[class]:"ERROR", 
	     level);

    user = group->users;
    while( user != NULL ) {
      logging_log_user( user, class, level, buffer, indata, datebuf, timebuf);
      user = user->next;
    }
  }
}

/*
 * LOUS debugging stuff
 */
#ifdef LOUS_DEBUG
s32_t LD_func;
s32_t LD_debug;
s32_t LL_trace;
s32_t LL_verbose;
s32_t LL_veryverbose;

loggroup_t* lous_loggroup;
loguser_t* lous_loguser;

#endif
