/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)alloc.c 1.3      04/02/21 Chris G. Davis"

/** \file src/memory/alloc.c
 *  \brief Memory Allocation Code
 *  \author Chris G. Davis 
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"

struct LOUS_MEMALLOC_Info;
typedef struct LOUS_MEMALLOC_Info lous_memalloc_t;

struct LOUS_MEMALLOC_Info {
  lous_memalloc_t* next;

  /*
   *  allocation info
   */
  s32_t alloc_count[127];
  void* alloc_addr[127][256];
  s32_t alloc_size[127][256];

};

lous_memalloc_t* lous_memalloc_list = NULL;
lous_memalloc_t* lous_memalloc_last = NULL;
s32_t lous_memalloc_count = 0;
FILE* lous_memalloc_out = NULL;
s32_t lous_memalloc_flags = 0;

#define lous_memalloc_hash(addr) (((unsigned int)addr >> 4 ) % 127)

void lous_memalloc_initialize(FILE* out, s32_t flags) {
  if( out == NULL ) {
    lous_memalloc_out = stderr;
  } else {
    lous_memalloc_out = out;
  }

  lous_memalloc_flags = flags;

}

lous_memalloc_t*  lous_memalloc_alloc() {
  lous_memalloc_t* retval;

  retval = malloc( sizeof( lous_memalloc_t ));
  memset( retval, 0, sizeof( lous_memalloc_t));

  return retval;
}

void lous_memalloc_add(void* ptr, s32_t size, 
		       char*fl, char* fn, s32_t ln) {
  s32_t index;
  s32_t cnt;
  s32_t found = -1;
  lous_memalloc_t* walker;

  if( lous_memalloc_list == NULL ) {
    lous_memalloc_list = lous_memalloc_alloc();
    lous_memalloc_last = lous_memalloc_list;
    lous_memalloc_count++;
  }

  walker = lous_memalloc_list;
  index = lous_memalloc_hash(ptr);

  while( (found == -1 ) && (walker != NULL )) {
    for(cnt=0; cnt < walker->alloc_count[index]; cnt++ ) {
      if( walker->alloc_addr[index][cnt] == ptr ) {
	found = cnt;
	break;
      }
    }
    if( found == -1 ) {
      walker = walker->next;
    }
  }
  
  /* found is -1 or # */
  if( found != -1 ) {
    /* found it */
    if( walker->alloc_size[index][found] != 0 ) {
      /* doh, already alloced */
      fprintf(lous_memalloc_out, "EEEK: alloc returned inuse addr: %p ",
	      ptr);
      fprintf(lous_memalloc_out, "Called from %s in %s, line %d\n",
	      fn, fl, ln);
    } else {
      /* do it */
      walker->alloc_size[index][found] = size;
    }
  } else {
    /* didn't find it */
    if( lous_memalloc_last->alloc_count[index] == 255 ) {
      /* we're full */
      lous_memalloc_last->next = lous_memalloc_alloc();
      lous_memalloc_count++;
      lous_memalloc_last = lous_memalloc_last->next;
    } 
    
    walker = lous_memalloc_last;
    walker->alloc_size[index][walker->alloc_count[index]] = size;
    walker->alloc_addr[index][walker->alloc_count[index]] = ptr;
    walker->alloc_count[index]++;

  }
}

s32_t lous_memalloc_del(void* ptr,
			char*fl, char* fn, s32_t ln) {
  s32_t retval = 0;
  s32_t index;
  s32_t cnt;
  s32_t found = -1;
  lous_memalloc_t* walker;

  walker = lous_memalloc_list;
  index = lous_memalloc_hash(ptr);

  while( ( found == -1 ) && (walker != NULL )) {
    for(cnt=0; cnt < walker->alloc_count[index]; cnt++ ) {
      if( walker->alloc_addr[index][cnt] == ptr ) {
	found = cnt;
	break;
      }
    }
    if( found == -1 ) {
      walker = walker->next;
    }
  }
  
  /* found is -1 or # */
  if( found != -1 ) {
    /* found it */
    if( walker->alloc_size[index][found] == 0 ) {
      /* doh, already deleted */
      fprintf(lous_memalloc_out, "EEEK: Free already free'd address addr: %p ",
	      ptr);
      fprintf(lous_memalloc_out, "Called from %s in %s, line %d\n",
	      fn, fl, ln);
    } else {
      /* do it */
      retval = walker->alloc_size[index][found];
      walker->alloc_size[index][found] = 0;
    }
  } else {
    /* oops, we didn't have this one */
    fprintf(lous_memalloc_out, "EEEK: tried to free unknown addr: %p ",
	    ptr);
    fprintf(lous_memalloc_out, "Called from %s in %s, line %d\n",
	    fn, fl, ln);
  }

  return retval;
}

void* lous_calloc(s32_t* count, s32_t nmemb, s32_t size,
		  char*fl, char* fn, s32_t ln) {
  void* retval = NULL;

  errno = 0;

  retval = calloc(nmemb, size);

  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
    fprintf(lous_memalloc_out, "CALLOC: %d bytes got %p [%s in %s line %d: %d => ",
	    size * nmemb, retval, fn, fl, ln, *count);
  }

  if( retval != NULL ) {
    (*count) += (nmemb * size);
    lous_memalloc_add(retval, nmemb * size, fl, fn, ln);
  } else {
    fprintf(lous_memalloc_out, "EEEK: calloc failed: %d bytes [%s in %s line %d]\n",
	    size * nmemb, fn, fl, ln);
  }

  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
    fprintf(lous_memalloc_out, "%d]\n", *count);
  }

  return retval;
}

void* lous_malloc(s32_t* count, s32_t size,
		  char*fl, char* fn, s32_t ln) {
  void* retval = NULL;

  errno = 0;

  retval = malloc(size);

  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
    fprintf(lous_memalloc_out, "ALLOC: %d bytes got %p [%s in %s line %d: %d => ",
	    size, retval, fn, fl, ln, *count);
  }

  if( retval != NULL ) {
    (*count) += size;
    lous_memalloc_add(retval, size, fl, fn, ln);
  } else {
    fprintf(lous_memalloc_out, "EEEK: malloc failed: %d bytes [%s in %s line %d]\n",
	    size, fn, fl, ln);
  }

  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
    fprintf(lous_memalloc_out, "%d]\n", *count);
  }

  return retval;
}

void lous_free(s32_t* count, void *ptr,
	       char*fl, char* fn, s32_t ln) {
  s32_t osize = 0;

  errno = 0;

  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_FREE ) {
    fprintf(lous_memalloc_out, "FREE: %p [%s in %s line %d: %d => ",
	    ptr, fn, fl, ln, *count);
  }

  osize = lous_memalloc_del(ptr, fl, fn, ln);
  (*count) -= osize;
  if( osize != 0 ) {
    free(ptr);
  }
  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_FREE ) {
    fprintf(lous_memalloc_out, "%d]\n", *count);
  }
}

void* lous_realloc(s32_t* count, void *ptr, s32_t size,
		   char*fl, char* fn, s32_t ln) {
  void* retval = NULL;
  s32_t osize = 0;
  
  errno = 0;
  
  retval = realloc(ptr, size);
  
  if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
    fprintf(lous_memalloc_out, "REALLOC: %p [%s in %s line %d: %d => ",
	    ptr, fn, fl, ln, *count);
  }
  
  if( retval != NULL ) {
    osize = lous_memalloc_del(retval, fl, fn, ln);
    (*count) -= osize;
    if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
      fprintf(lous_memalloc_out, "%d] %d -> %d => %p\n",
	      *count, osize, size, retval);
    }
    lous_memalloc_add(retval, size, fl, fn, ln);
    (*count) += osize;
  } else {
    if( lous_memalloc_flags & LOUS_MEMALLOC_SHOW_ALLOC ) {
      fprintf(lous_memalloc_out, " NULL]\n");
    }
  }    
  
  return retval;
}

void lous_memalloc_func_state(s32_t* count, char* fl, char* fn, s32_t ln) {
  fprintf(lous_memalloc_out, "Func: [%s in %s line %d: %d]\n", 
	  fn, fl, ln, *count);
}

void lous_memalloc_state() {
  lous_memalloc_t* walker;
  s32_t count = 0;
  s32_t index;
  s32_t ent;
  
  fprintf(lous_memalloc_out, "Count: %d Flags: %d\n",
	  lous_memalloc_count, lous_memalloc_flags);
  walker = lous_memalloc_list;

  while( walker != NULL ) {
    /* look for outstanding allocations */
    fprintf(lous_memalloc_out, "Oustanding Allocations: %d [%p]\n", 
	    count, walker);
    fprintf(lous_memalloc_out, "==============================================\n");

    for(index = 0; index < 127; index++) {
      for(ent = 0; ent < walker->alloc_count[index]; ent++) {
	if( walker->alloc_size[index][ent] != 0 ) {
	  fprintf(lous_memalloc_out, "  %p = %d\n",
		  walker->alloc_addr[index][ent],
		  walker->alloc_size[index][ent]);
	}
      }
    }

    walker = walker->next;
  }

}
