/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)iproute.c 1.5      04/02/21 Chris G. Davis"

/** \file src/networking/iproute.c
 *  \brief Network Interface Code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <netinet/ip.h>
#include <errno.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/if_tunnel.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/if_arp.h>
#include <time.h>

#include "lous/lous.h"
#include "lous/iproute.h"

struct idxmap
{
  struct idxmap * next;
  int             index;
  char            name[16];
};

static struct idxmap *idxmap = NULL;

char* findMinNetmask(char* ip, char* gip) {
  u32_t lip[4] = {0,0,0,0};
  u32_t fip[4] = {0,0,0,0};
  u32_t tip[4] = {0,0,0,0};
  u32_t nm[4] = {0,0,0,0};
  char* nmbuffer;
  int i,j;

  sscanf(ip, "%u.%u.%u.%u", &lip[0], &lip[1], &lip[2], &lip[3]);
  sscanf(gip, "%u.%u.%u.%u", &fip[0], &fip[1], &fip[2], &fip[3]);
  
  for(i = 0; i < 4; i++ ) {
    tip[i] = ~( lip[i] ^ fip[i] );
    
    for(j=7; j>=0; j--) {
      if( tip[i] & (1 << j) )
        nm[i] |= 1 << j;
      else
        goto found_mask;
    }
  }
 found_mask:
  nmbuffer = malloc(16);
  sprintf(nmbuffer,"%u.%u.%u.%u", nm[0], nm[1], nm[2], nm[3]);
  
  return nmbuffer;
}

/*
 * most of this code is based on iproute2, by Alexey K., but presented here in a slightly
 * more logical form
 *
 * It's also much less featurful.
 */
int iproute_tunnel(int cmd, int encaps, char* tname, struct in_addr* local, struct in_addr* remote) {
  struct ip_tunnel_parm tp;
  struct ifreq intf;
  int fd;

  memset(&tp, 0, sizeof( struct ip_tunnel_parm ));
  tp.iph.version = 4;
  tp.iph.ihl = 5;
  tp.iph.frag_off = htons(IP_DF);
  tp.iph.ttl = 64;

  strncpy(tp.name, tname, IFNAMSIZ);

  switch(cmd) {
  case SIOCADDTUNNEL:
    if( encaps == IPPROTO_GRE ) {
      /* FIXME: we don't really handle GRE yet */
      strncpy(intf.ifr_name, "gre0", IFNAMSIZ);
    } else {
      strncpy(intf.ifr_name, "tunl0", IFNAMSIZ);
    }

    tp.iph.protocol = encaps;
    tp.iph.daddr = remote->s_addr;
    tp.iph.saddr = local->s_addr;
   
    break;
  case SIOCDELTUNNEL:
    strncpy(intf.ifr_name, tname, IFNAMSIZ);
    break;
  default:
    return -1;
    break;
  }

  intf.ifr_ifru.ifru_data = (void*)&tp;

  if( (fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
    return -1;
  }

  if( ioctl(fd, cmd, &intf) != 0 ) {
    close(fd);
    return -1;
  }

  close(fd);

  return 0;
}


int iproute_tunnel_add(int encaps, char* tname, struct in_addr* local, struct in_addr* remote) {
  return iproute_tunnel(SIOCADDTUNNEL, encaps, tname, local, remote);
}

int iproute_tunnel_del(char* tname) {
  return iproute_tunnel(SIOCDELTUNNEL, 0, tname, NULL, NULL);
}

int iproute_addr_cmd(int cmd, char* ifname, struct in_addr* addr) {
  struct rtnl_handle rth;

  struct {
    struct nlmsghdr n;
    struct ifaddrmsg ifa;
    char   buf[256];
  } req;

  req.n.nlmsg_len = NLMSG_LENGTH(sizeof( struct ifaddrmsg ));
  req.n.nlmsg_flags = NLM_F_REQUEST;
  req.n.nlmsg_type = cmd;
  req.ifa.ifa_family = AF_INET;
  req.ifa.ifa_prefixlen = 32; /* for mobile IP, we always deal with /32 */

  rtnl_addattr_l(&req.n, sizeof(req), IFA_LOCAL, (void*)addr, sizeof( struct in_addr));

  if( rtnl_open(&rth, 0) != 0 ) {
    return -1;
  }

  if( ll_init_map(&rth) != 0 ) {
    close(rth.fd);
    return -1;
  }

  if( (req.ifa.ifa_index = ll_name_to_index(ifname)) == 0 ) {
    ll_free_map();
    close(rth.fd);
    return -1;
  }

  ll_free_map();

  if(rtnl_talk(&rth, &req.n, NULL) < 0 ) {
    close(rth.fd);
    return -1;
  }

  close(rth.fd);
  return 0;
}

int iproute_addr_add(char* ifname, struct in_addr* addr) {
  return iproute_addr_cmd(RTM_NEWADDR, ifname, addr);
}

int iproute_addr_del(char* ifname, struct in_addr* addr) {
  return iproute_addr_cmd(RTM_DELADDR, ifname, addr);
}

int iproute_link_set(char* ifname, int flag, int value) {
  struct ifreq intf;
  int fd;

  strncpy(intf.ifr_name, ifname, IFNAMSIZ);

  if( (fd = socket(AF_INET, SOCK_DGRAM, 0 )) < 0 ) {
    return -1;
  }

  if( ioctl(fd, SIOCGIFFLAGS, &intf) != 0 ) {
    close(fd);
    return -1;
  }

  if( value ) {
    intf.ifr_flags |= flag;
  } else {
    intf.ifr_flags &= ~flag;
  }

  if( ioctl(fd, SIOCSIFFLAGS, &intf) != 0 ) {
    close(fd);
    return -1;
  }

  close(fd);

  return 0;
}

int iproute_link_set_up(char* ifname) {
  return iproute_link_set(ifname, IFF_UP, 1);
}

int iproute_link_set_down(char* ifname) {
  return iproute_link_set(ifname, IFF_UP, 0);
}

int iproute_route(int cmd, int flags, int type, char* dev, struct in_addr* dest, int cidr, int table) {
  struct rtnl_handle rth;
  struct {
    struct nlmsghdr n;
    struct rtmsg r;
    char buf[1024];
  } req;
  int idx;
  
  memset(&req, 0, sizeof(req));
  req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
  req.n.nlmsg_flags = flags | NLM_F_REQUEST;
  req.n.nlmsg_type = cmd;
  req.r.rtm_table = RT_TABLE_MAIN;
  fprintf(stderr, "route do: %d %d %d %d on %s, %X/%d\n", cmd, flags,  req.n.nlmsg_flags, type, 
	dev, dest->s_addr, cidr);  

  if (cmd == RTM_DELROUTE) {
    req.r.rtm_protocol = RTPROT_UNSPEC;
    req.r.rtm_scope = RT_SCOPE_NOWHERE;
  } else {
    req.r.rtm_protocol = RTPROT_BOOT;
    req.r.rtm_scope = RT_SCOPE_LINK;
  }

  req.r.rtm_type = type;
  req.r.rtm_family = AF_INET;
  
  if (dest != NULL) {
    rtnl_addattr_l(&req.n, sizeof(req), RTA_DST, (void *) dest, 4);
    req.r.rtm_dst_len = cidr;
  }

  if ( table != -1 ) {
	req.r.rtm_table = table;
  }

  if (rtnl_open(&rth, 0) != 0) {
    fprintf(stderr, "rtnl_open failed: %d\n", errno);
    return -1;
  }
  
  if (dev != NULL) {
    if (ll_init_map(&rth) != 0) {
      fprintf(stderr, "ll_init failed: %d\n", errno);
      ll_free_map();
      close(rth.fd);
      return -1;
    }
    
    if ((idx = ll_name_to_index((char*)dev)) == 0) {
      fprintf(stderr, "ll_n2i failed: %d\n", errno);
      ll_free_map();
      close(rth.fd);
      return -1;
    }
    fprintf(stderr, "Found %s at %d\n", dev, idx);
    rtnl_addattr32(&req.n, sizeof(req), RTA_OIF, idx);
    ll_free_map();
  }
  
  if (rtnl_talk(&rth, &req.n, NULL) < 0) {
    fprintf(stderr, "rtnl_talk failed: %d\n", errno);
    close(rth.fd);
    return -1;
  }
  
  close(rth.fd);
  return 0;
  
}

int iproute_route_add(char* dev, struct in_addr* dest, int table) {
  return iproute_route(RTM_NEWROUTE, NLM_F_CREATE, RTN_UNICAST, dev, dest, 32, table);

}

int iproute_route_del(char* dev, struct in_addr* dest, int table) {
  return iproute_route(RTM_DELROUTE, 0, RTN_UNSPEC, dev, dest, 32, table);
}

int rtnl_open(struct rtnl_handle *rth, unsigned subscriptions) {
  int addr_len;
  
  memset(rth, 0, sizeof(rth));
  
  rth->fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  if (rth->fd < 0) {
    fprintf(stderr, "socket failed: %d\n", errno);
    return -1;
  }
  
  memset(&rth->local, 0, sizeof(rth->local));
  rth->local.nl_family = AF_NETLINK;
  rth->local.nl_groups = subscriptions;
  
  if (bind(rth->fd, (struct sockaddr*)&rth->local, sizeof(rth->local)) < 0) {
    fprintf(stderr, "bind failed: %d\n", errno);
    return -1;
  }
  addr_len = sizeof(rth->local);
  if (getsockname(rth->fd, (struct sockaddr*)&rth->local, &addr_len) < 0) {
    fprintf(stderr, "getsockname failed: %d\n", errno);
    return -1;
  }
  if (addr_len != sizeof(rth->local)) {
    fprintf(stderr, "Wrong address length %d\n", addr_len);
    return -1;
  }
  if (rth->local.nl_family != AF_NETLINK) {
    fprintf(stderr, "Wrong address family %d\n", rth->local.nl_family);
    return -1;
  }
  rth->seq = time(NULL);
  return 0;
}

void rtnl_close(struct rtnl_handle *rth) {
  close(rth->fd);
}

int rtnl_talk(struct rtnl_handle *rth, struct nlmsghdr *n, struct nlmsghdr *answer) {
        int status;
        struct nlmsghdr *h;
        struct sockaddr_nl nladdr;
        struct iovec iov;
        char   buf[8192];
        struct msghdr msg = {
                (void *) &nladdr, sizeof(nladdr),
                &iov, 1,
                NULL, 0,
                0
        };

        iov.iov_base = (void*) n;
        iov.iov_len = n->nlmsg_len;

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;
        nladdr.nl_pid = 0;
        nladdr.nl_groups = 0;

        n->nlmsg_seq = ++rth->seq;

        if (answer == NULL)
                n->nlmsg_flags |= NLM_F_ACK;

        status = sendmsg(rth->fd, &msg, 0);

        if (status < 0) {
                fprintf(stderr, "rtnl_talk: sendmsg: %s\n", strerror(errno));
                return -1;
        }

        iov.iov_base = buf;
        iov.iov_len = sizeof(buf);

        do {
                status = recvmsg(rth->fd, &msg, 0);

                if (status < 0) {
                        if (errno == EINTR)
                                continue;
                        fprintf(stderr, "rtnl_talk: OVERRUN: %s\n",
                                         strerror(errno));
                        continue;
                }
                if (status == 0) {
                        fprintf(stderr, "rtnl_talk: EOF on netlink\n");
                        return -1;
                }
                if (msg.msg_namelen != sizeof(nladdr)) {
                        fprintf(stderr,
                              "rtnl_talk: sender address length == %i != %i\n",
                              msg.msg_namelen, sizeof(nladdr));
                        return -1;
                }
                for (h = (struct nlmsghdr*) buf; status >= sizeof(*h); ) {
                        int len = h->nlmsg_len;
                        pid_t pid = h->nlmsg_pid;
                        int l = len - sizeof(*h);
                        unsigned seq = h->nlmsg_seq;
                        if (l < 0 || len > status) {
                                if (msg.msg_flags & MSG_TRUNC) {
                                        fprintf(stderr,
                                              "rtnl_talk: "
                                              "truncated message\n");
                                        return -1;
                                }
                                fprintf(stderr,
                                      "rtnl_talk: malformed message: len=%d\n",
                                      len);
                                return -1;
                        }

                        if (h->nlmsg_pid != pid || h->nlmsg_seq != seq) {
                                /* junk */
                                continue;
                        }

                        if (h->nlmsg_type == NLMSG_ERROR) {
                                struct nlmsgerr *err =
                                        (struct nlmsgerr *) NLMSG_DATA(h);
                                if (l < sizeof(struct nlmsgerr)) {
                                        fprintf(stderr,
                                              "rtnl_talk: ERROR truncated\n");
                                        return -1;
                                }

                                errno = -err->error;
                                if (errno == 0) {
                                        if (answer)
                                                memcpy(answer, h,
                                                       h->nlmsg_len);
                                        return 0;
                                }
                                fprintf(stderr,
                                      "rtnl_talk: RTNETLINK error: "
                                      "%s\n", strerror(errno));

                                return -1;
                        }

                        if (answer) {
                                memcpy(answer, h, h->nlmsg_len);
                                return 0;
                        }

                        fprintf(stderr, "rtnl_talk: unexpected reply\n");

                        status -= NLMSG_ALIGN(len);
                        h = (struct nlmsghdr *)
                                (((char *) h) + NLMSG_ALIGN(len));
                }
                if (msg.msg_flags & MSG_TRUNC) {
                        fprintf(stderr, "rtnl_talk: message truncated\n");
                        continue;
                }
                if (status) {
                        fprintf(stderr, "rtnl_talk: remnant of size %i\n",
                              status);
                        return -1;
                }
        } while (1);

}

int rtnl_send(struct rtnl_handle *rth, char *buf, int len) {
  struct sockaddr_nl nladdr;
  
  memset(&nladdr, 0, sizeof(nladdr));
  nladdr.nl_family = AF_NETLINK;
  
  return sendto(rth->fd, buf, len, 0, (struct sockaddr*)&nladdr, sizeof(nladdr));
}

int rtnl_wilddump_request(struct rtnl_handle *rth, int family, int type) {
  struct {
    struct nlmsghdr nlh;
    struct rtgenmsg g;
  } req;
  struct sockaddr_nl nladdr;

  memset(&nladdr, 0, sizeof(nladdr));
  nladdr.nl_family = AF_NETLINK;

  req.nlh.nlmsg_len = sizeof(req);
  req.nlh.nlmsg_type = type;
  req.nlh.nlmsg_flags = NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
  req.nlh.nlmsg_pid = 0;
  req.nlh.nlmsg_seq = rth->dump = ++rth->seq;
  req.g.rtgen_family = family;

  return sendto(rth->fd, (void*)&req, sizeof(req), 0, (struct sockaddr*)&nladdr, sizeof(nladdr));
}

int rtnl_dump_filter(struct rtnl_handle *rth,
                     int (*filter)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
                     void *arg1) {
  char    buf[8192];
  struct sockaddr_nl nladdr;
  struct iovec iov = { buf, sizeof(buf) };

  while (1) {
    int status;
    struct nlmsghdr *h;

    struct msghdr msg = {
      (void*)&nladdr, sizeof(nladdr),
      &iov,   1,
      NULL,   0,
                        0
    };

    status = recvmsg(rth->fd, &msg, 0);

    if (status < 0) {
      if (errno == EINTR)
	continue;
      continue;
    }
    if (status == 0) {
      return -1;
    }
    if (msg.msg_namelen != sizeof(nladdr)) {
      exit(1);
    }

    h = (struct nlmsghdr*)buf;
    while (NLMSG_OK(h, status)) {
      int err;

      if (h->nlmsg_pid != rth->local.nl_pid ||
	  h->nlmsg_seq != rth->dump) {
	goto skip_it;
      }

      if (h->nlmsg_type == NLMSG_DONE)
	return 0;
      if (h->nlmsg_type == NLMSG_ERROR) {
	struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(h);
	if (h->nlmsg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr))) {
	} else {
	  errno = -err->error;
	}
	return -1;
      }
      err = filter(&nladdr, h, arg1);
      if (err < 0)
	return err;

    skip_it:
      h = NLMSG_NEXT(h, status);
    }
    if (msg.msg_flags & MSG_TRUNC) {
      continue;
    }
  }
}

int rtnl_addattr32(struct nlmsghdr *n, int maxlen, int type, __u32 data) {
  int len = RTA_LENGTH(4);
  struct rtattr *rta;
  if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
    return -1;
  rta = (struct rtattr*)(((char*)n) + NLMSG_ALIGN(n->nlmsg_len));
  rta->rta_type = type;
  rta->rta_len = len;
  memcpy(RTA_DATA(rta), &data, 4);
  n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
  return 0;
}

int rtnl_addattr_l(struct nlmsghdr *n, int maxlen, int type, void *data, int alen) {
  int len = RTA_LENGTH(alen);
  struct rtattr *rta;
  
  if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
    return -1;
  rta = (struct rtattr*)(((char*)n) + NLMSG_ALIGN(n->nlmsg_len));
  rta->rta_type = type;
  rta->rta_len = len;
  memcpy(RTA_DATA(rta), data, alen);
  n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
  return 0;
}

int ll_init_map(struct rtnl_handle *rth) {
  if (rtnl_wilddump_request(rth, AF_UNSPEC, RTM_GETLINK) < 0) {
	fprintf(stderr, "cannot send dump");
	return -1;
  }
  
  if (rtnl_dump_filter(rth, ll_remember_index, &idxmap) < 0) {
    fprintf(stderr, "Dump terminated\n");
	return -1;
  }
  return 0;
}

int ll_name_to_index(char *name) {
  struct idxmap *im;

  if( name == NULL ) {
	return 0;
  }

  for( im = idxmap; im != NULL; im = im->next) {
	if( strcmp(im->name, name) == 0 ) 
	  return im->index;
  }

  return 0;
}

int ll_free_map() {
  struct idxmap *idx, *prev;

  idx=idxmap;
  while( idx != NULL ) {
	prev = idx;
	idx=idx->next;
	free(prev);
  }
  idxmap = NULL;

  return 0;
}

int ll_remember_index(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg) {
        struct idxmap **map = (struct idxmap **) arg;
        struct ifinfomsg *ifi = NLMSG_DATA(n);
        struct idxmap *im;
        struct rtattr *tb[IFLA_MAX + 1];

        if (map == NULL)
                map = &idxmap;

        if (n->nlmsg_type != RTM_NEWLINK)
                return 0;

        if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifi)))
                return -1;

        memset(tb, 0, sizeof(tb));
        rtnl_parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), IFLA_PAYLOAD(n));
        if (tb[IFLA_IFNAME] == NULL)
                return 0;

        im = malloc(sizeof(*im));
        if (im == NULL)
                return -1;

        im->next = *map;
        im->index = ifi->ifi_index;
        strncpy(im->name, RTA_DATA(tb[IFLA_IFNAME]),sizeof(im->name));
        *map = im;
        return 0;
}

int rtnl_parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta, int len) {
  while (RTA_OK(rta, len)) {
    if (rta->rta_type <= max)
      tb[rta->rta_type] = rta;
    rta = RTA_NEXT(rta,len);
  }
  return 0;
}
