/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)tokenizer.c 1.2      04/02/21 Chris G. Davis"

/** \file src/string/tokenizer.c
 *  \brief Tokenizer code
 *  \author Chris G. Davis 
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "lous/tokenizer.h"
#include "internal/tokenizer.h"

/*
  ok, we alloc a tokenizer_t, set the params

  then we feed it a string, and the tokenizer spits back a
  tokenizer_response_t, with a pile of cool stuff, like
  all the tokens separated, with strings up to and including 
  that point, and the remainder tokens

*/

tokenizer_t* tokenizer_alloc() {
  tokenizer_t* retval;

  LDFunc();
  
  Lmalloc( retval, sizeof( tokenizer_t ), NULL);

  return retval;
}

void tokenizer_destroy(tokenizer_t* izer) {
  LDFunc();
  LDParam(izer, == NULL, "%p");
  
  free( izer );
}

void tokenizer_expand_array( char** array, s32_t num) {
  char* old = NULL;

  LDFunc();

  if( *array ) {
    old = *array;
  }

  Lmalloc( *array, num, );

  if( old != NULL ) {
    memcpy(*array, old, num - 1);
  }
}

void tokenizer_add_separator(tokenizer_t* izer, char c) {
  LDFunc();
  LDParam(izer, == NULL, "%p");

  tokenizer_expand_array( &izer->separators, izer->num_separators + 1);
  izer->separators[izer->num_separators] = c;

  izer->num_separators++;
  
}

void tokenizer_add_terminator(tokenizer_t* izer, char c) {
  LDFunc();
  LDParam(izer, == NULL, "%p");

  tokenizer_expand_array( &izer->terminators, izer->num_terminators + 1);
  izer->terminators[izer->num_terminators] = c;

  izer->num_terminators++;
}

void tokenizer_add_ignored(tokenizer_t* izer, char c) {
  LDFunc();
  LDParam(izer, == NULL, "%p");

  tokenizer_expand_array( &izer->ignored, izer->num_ignored + 1);
  izer->ignored[izer->num_ignored] = c;

  izer->num_ignored++;
}

bool_t tokenizer_ischar(char c, char* list, s32_t num) {
  s32_t index;

  for(index = 0; index < num; index++ ) {
    if( c == list[index] ) {
      return true;
    }
  }

  return false;
}

/*
 * Mode of operation
 *
 * Parse the string, separating tokens, builds strings and remainders
 *  finish is when we get to end of string (0 terminated)
 *
 */
#define TOKENIZER_ISCHAR(ch, type) \
tokenizer_ischar(ch, izer->type, izer->num_##type)

s32_t tokenizer_do_one_token(tokenizer_t* izer,
			     char* str,
			     tokenizer_response_t* resp,
			     char* tbuffer) {
  s32_t retval = -1;

  char* remainder = NULL;

  char* walker = str;
  char* backone = NULL;
  char* start = NULL;
  char* end = NULL;
  char* igend = NULL;

  tokenizer_ent_t* ent;

  LDFunc();
  LDParam( str, == NULL, "%p", -1);
  LDParam( tbuffer, == NULL, "%p", -1);

  Lmalloc(ent, sizeof(tokenizer_ent_t), -1);

  /* extract token:
     find start of token (find first non-term, non-ignored, non-sep char)
     then find end of token
  */
  while( (*walker) && (end == NULL) && ( resp->got_terminator == false )) {

    if( TOKENIZER_ISCHAR(*walker, separators) == true ) {
      /*
	separators end a token. note that if the igend is set
	and we hit a separator, we use igend, not walker, since we ignore
	the ignored chars
      */
      if(start != NULL ) {
        resp->is_partial = false;
	if( igend != NULL ) {
	  end = igend;
	} else {
	  end = backone;
	}
      }
    } else if( TOKENIZER_ISCHAR(*walker, terminators) == true ) {
      /*
	so, we got a terminator. end the token, and end the tokenizing
      */
      if(start != NULL ) {
	if( igend != NULL ) {
	  end = igend;
	} else {
	  end = backone;
	}
      }

      resp->got_terminator = true;

    } else if( TOKENIZER_ISCHAR(*walker, ignored) == true ) {
      /* 
	 For ignored chars, we don't include them in the token, unless we need to
      */
      if( start != NULL ) {
	igend = backone;
      }
    } else {
      /* we got a different char */
      if( start == NULL ) {
	start = walker;
        resp->is_partial = true;
      }

      igend = NULL;
    }

    backone = walker;
    walker++;
  }

  if( start != NULL ) {
    if( end == NULL ) {
      end = backone;
    }


    if( end != NULL ) {
      /* got a token */
      if( strlen( walker ) > 0 ) {
	/* build remainder */
	Lstrdup( remainder, walker, -1);
      }
      ent->remainder = remainder;
      
      /* make zero terminated token */
      Lmalloc(ent->token, (end - start + 2), -1, 
	      free( ent->remainder); free( ent ); );
      
      memcpy(ent->token, start, (end - start + 1));
      
      /* make string */
      if( tbuffer[0] != 0 ) {
	strncat(tbuffer, izer->separators, 1);
      }
      
      Lstrdup(ent->preceeding, tbuffer, -1, 
	      free( ent->remainder); free( ent->token); free( ent ); );
      strcat(tbuffer, ent->token);
      Lstrdup(ent->string, tbuffer, -1, 
	      free( ent->remainder); free( ent->token); free( ent->preceeding ); free( ent ); );
      
      /* add stuff to response */
      ll_insert_end(resp->tokens, ent);
      resp->num_tokens++;
      
      /* recurse */
      if( (remainder != NULL) && ( resp->got_terminator == false ) ) {
	retval = tokenizer_do_one_token(izer, remainder, resp, tbuffer);
      } else {
	retval = 0;
      }
    }
  } else {
    retval = 0;
  }
  
  return retval;
}

tokenizer_response_t* tokenizer_tokenize(tokenizer_t* izer, 
					  char* str) {
  tokenizer_response_t* retval;
  char* tbuffer;

  LDFunc();
  LDParam(izer, == NULL, "%p", NULL);
  LDParam(str, == NULL, "%p", NULL);

  Lmalloc(retval, sizeof( tokenizer_response_t ), NULL);
  Lmalloc(tbuffer, strlen( str ), NULL, free( retval ); );

  if( (retval->tokens = ll_alloc(0, NULL)) == NULL ) {
    LError("Unable to alloc linkedlist");
    free( tbuffer );
    free( retval );
    return NULL;
  }

  if( tokenizer_do_one_token(izer, str, retval, tbuffer) !=0 ) {
    tokenizer_destroy_response( retval );
    return NULL;
  }

  return retval;
}

void tokenizer_destroy_response(tokenizer_response_t* resp) {
  tokenizer_ent_t* ent;

  LDFunc();
  LDParam(resp, == NULL, "%p");

  if( resp->tokens != NULL ) {
    while( ( ent = ll_remove_begin(resp->tokens)) != NULL ) {
      if( ent->preceeding ) {
	free( ent->preceeding );
      }
      
      if( ent->string ) {
	free( ent->string );
      }
      
      if( ent->token ) {
	free( ent->token );
      }
      
      if( ent->remainder ) {
	free( ent->remainder );
      }
      
      free( ent );
    }

    ll_destroy( resp->tokens );
  }

  free( resp );

}

s32_t tokenizer_get_numtokens(tokenizer_response_t* resp) {
  LDFunc();
  LDParam(resp, == NULL, "%p", -1);

  return resp->num_tokens;
}

bool_t tokenizer_get_terminator(tokenizer_response_t* resp) {
  LDFunc();
  LDParam(resp, == NULL, "%p", -1);

  return resp->got_terminator;
}

bool_t tokenizer_get_ispartial(tokenizer_response_t* resp) {
  LDFunc();
  LDParam(resp, == NULL, "%p", -1);

  return resp->is_partial;
}

char* tokenizer_get_token(tokenizer_response_t* resp, s32_t index) {
  s32_t count;

  LDFunc();
  LDParam(resp, == NULL, "%p", NULL);
  LDParam(resp->tokens, == NULL, "%p", NULL);
  LDParam(index, >= resp->num_tokens, "%d", NULL);

  ll_clearnext( resp->tokens );
  for(count = 0; count < index; count++ ) {
    ll_getnext( resp->tokens );
  }

  return ((tokenizer_ent_t*)ll_getnext( resp->tokens ))->token;
}

char* tokenizer_get_preceeding(tokenizer_response_t* resp, s32_t index) {
  s32_t count;

  LDFunc();
  LDParam(resp, == NULL, "%p", NULL);
  LDParam(resp->tokens, == NULL, "%p", NULL);
  LDParam(index, >= resp->num_tokens, "%d", NULL);

  ll_clearnext( resp->tokens );
  for(count = 0; count < index; count++ ) {
    ll_getnext( resp->tokens );
  }

  return ((tokenizer_ent_t*)ll_getnext( resp->tokens ))->preceeding;
}

char* tokenizer_get_string(tokenizer_response_t* resp, s32_t index) {
  s32_t count;

  LDFunc();
  LDParam(resp, == NULL, "%p", NULL);
  LDParam(resp->tokens, == NULL, "%p", NULL);
  LDParam(index, >= resp->num_tokens, "%d", NULL);

  ll_clearnext( resp->tokens );
  for(count = 0; count < index; count++ ) {
    ll_getnext( resp->tokens );
  }

  return ((tokenizer_ent_t*)ll_getnext( resp->tokens ))->string;
}

char* tokenizer_get_remainder(tokenizer_response_t* resp, s32_t index) {
  s32_t count;

  LDFunc();
  LDParam(resp, == NULL, "%p", NULL);
  LDParam(resp->tokens, == NULL, "%p", NULL);
  LDParam(index, >= resp->num_tokens, "%d", NULL);

  ll_clearnext( resp->tokens );
  for(count = 0; count < index; count++ ) {
    ll_getnext( resp->tokens );
  }

  return ((tokenizer_ent_t*)ll_getnext( resp->tokens ))->remainder;
}


