/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)hash-default.c 1.6      04/02/16 Chris G. Davis"

/** \file src/structures/hash-default.c
 *  \brief Hash table default routines
 *  \author Chris G. Davis
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "lous/hash.h"
#include "internal/hash.h"

void hash_delete_entry(hash_ent_t* ent, void (*free_func)(void*)) {

  if( ent->key ) {
    free( ent->key );
    ent->key = NULL;
  }

  if( ent->data ) {
    if( free_func ) {
      free_func( ent->data );
    }
    ent->data = NULL;
  }

  free(ent);
}

s32_t hash_default_initialize(hash_t* hash) {
  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  Lmalloc( hash->entries, hash->buckets * sizeof( hash_ent_t), -1);

  return 0;
}

s32_t hash_default_destroy(hash_t* hash) {
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  for(index = 0; index < hash->buckets; index++) {
    if( hash->entries[index] != NULL ) {
      hash_delete_entry( hash->entries[index], hash->free_func);
    }
  }

  free( hash->entries );

  return 0;
}

s32_t hash_default_insert(hash_t* hash, void* key, void* data) {
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);
  LDParam(data, != data, "%p", -1);

  /* can we do this */
  index = hash->hash_func(key, hash->keylen, hash->buckets);
  LDebug(LD_debug, LL_trace, "got index %d", index);

  if( index >= 0 ) {
    if( hash->entries[index] != NULL ) {
      errno = EEXIST;
    } else {
      /* initialize the new entry */
      Lmalloc( hash->entries[index], sizeof( hash_ent_t), -1 );

      if( hash->keylen ) {
	Lmalloc( hash->entries[index]->key,  hash->keylen, -1, free(hash->entries[index]); );
      } else {
	Lstrdup( hash->entries[index]->key, key, -1, free(hash->entries[index]); );
      }
      
      hash->entries[index]->data = data;
      hash->num_entries++;

      if( hash->flags & LIST_ITERATOR ) {
	iterator_insert(hash, data );
      }
    }
  }
  
  return 0;
}

void* hash_default_remove(hash_t* hash, void* key) {
  void* retval = NULL;
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);

  index = hash->hash_func(key, hash->keylen, hash->buckets);
  if( index >= 0 ) {
    if( hash->entries[index] == NULL ) {
      errno = EEXIST;
    } else {
      retval = hash->entries[index]->data;
      if( hash->flags & LIST_ITERATOR ) {
	iterator_remove(hash, retval);
      }
      hash_delete_entry(hash->entries[index], NULL );
      hash->entries[index] = NULL;
      hash->num_entries--;
    }
  }

  return retval;
}

s32_t hash_default_delete(hash_t* hash, void* key) {
  s32_t retval = -1;
  void* odata;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);

  odata = hash_default_remove(hash, key);

  if( odata != NULL ) {
    if( hash->free_func != NULL ) {
      hash->free_func( odata );
    }
    retval = 0;
  }

  return retval;
}

void* hash_default_replace(hash_t* hash, void* key, void* data) {
  void* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);
  LDParam(data, != data, "%p", NULL);

  retval = hash_default_remove(hash, key);
  hash_default_insert(hash, key, data);

  return retval;
}

void* hash_default_lookup(hash_t* hash, void* key) {
  void* retval = NULL;
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);

  index = hash->hash_func(key, hash->keylen, hash->buckets);
  LDebug(LD_debug, LL_trace, "got index %d", index);

  if( index >= 0 ) {
    if( hash->entries[index] == NULL ) {
      errno = EEXIST;
    } else {
      retval = hash->entries[index]->data;
    }
  }

  return retval;
}

void** hash_default_getkeys(hash_t* hash) {
  void** retval = NULL;
  s32_t index;
  s32_t keynum = 0;
 
  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);

  Lmalloc( retval, (hash->num_entries + 1) * sizeof( void*), NULL);

  for(index = 0; index < hash->buckets; index++) {
    if( hash->entries[index] != NULL ) {
      if( hash->keylen ) {
	Lmalloc( retval[keynum], hash->keylen, NULL, free(retval); );
	memcpy( retval[keynum], hash->entries[index]->key, hash->keylen );
      } else {
	Lstrdup(retval[keynum], hash->entries[index]->key, NULL, free( retval ); );
      }
      keynum++;
    }
  }

  return retval;
}

s32_t hash_default_numentries(hash_t* hash) {
  s32_t retval = -1;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  retval = hash->num_entries;

  return retval;
}

hash_ops_t generic_ops = {
  .initialize = hash_default_initialize,
  .destroy = hash_default_destroy,
  .insert = hash_default_insert,
  .remove = hash_default_remove,
  .delete = hash_default_delete,
  .replace = hash_default_replace,
  .lookup = hash_default_lookup,
  .getkeys = hash_default_getkeys,
  .numentries = hash_default_numentries,
};

