/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)hash-splay.c 1.6      04/02/16 Chris G. Davis"

/** \file src/structures/hash-splay.c
 *  \brief Hash table splay routines
 *  \author Chris G. Davis
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/lous.h"
#include "lous/logging.h"
#include "lous/hash.h"
#include "lous/splay.h"
#include "internal/hash.h"
#include "internal/splay.h"

hash_ent_t* hash_splay_alloc_ent(hash_t* hash) {
  hash_ent_t* retval = NULL;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);

  Lmalloc( retval, sizeof( hash_ent_t ), NULL);

  retval->key = NULL;
  retval->data = splay_alloc(0, hash->keylen, hash->comp_func, hash->free_func);

  if( retval->data == NULL ) {
    LError("Unable to alloc splay entry");
    free( retval );
    retval = NULL;
  }
  
  return retval;
}

s32_t hash_splay_destroy(hash_t* hash) {
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  for(index= 0; index < hash->buckets; index++) {
    if( hash->entries[index] != NULL ) {
      splay_destroy( hash->entries[index]->data );
      free( hash->entries[index] );
    }
  }

  free( hash );

  return 0;
}

s32_t hash_splay_initialize(hash_t* hash) {
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  
  Lmalloc(hash->entries, hash->buckets * sizeof( hash_ent_t*), -1);

  for(index= 0; index < hash->buckets; index++) {
    hash->entries[index] = hash_splay_alloc_ent(hash);
    
    if( hash->entries[index] == NULL ) {
      hash_splay_destroy(hash);
      errno = ENOMEM;
      return -1;
    }

  }

  return 0;
}

s32_t hash_splay_insert(hash_t* hash, void* key, void* data) {
  s32_t retval = -1;
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);
  LDParam(data, != data, "%p", -1);

  /* can we do this */
  index = hash->hash_func(key, hash->keylen, hash->buckets);

  if( index < 0 ) {
    errno = EINVAL;
  } else {
    retval = splay_insert(hash->entries[index]->data, key, data);

    if( hash->flags & LIST_ITERATOR ) {
      iterator_insert(hash, data );
    }
  }

  return retval;
}

void* hash_splay_remove(hash_t* hash, void* key) {
  void* retval = NULL;
  s32_t index;

  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);
  
  /* can we do this */
  index = hash->hash_func(key, hash->keylen, hash->buckets);

  if( index < 0 ) {
    errno = EINVAL;
  } else {
    retval = splay_remove(hash->entries[index]->data, key);

    if( hash->flags & LIST_ITERATOR ) {
      iterator_remove(hash, retval);
    }
  }

  return retval;
}

s32_t hash_splay_delete(hash_t* hash, void* key) {
  s32_t retval = -1;
  void* odata;

  LDFunc();
  LDParam(hash, == NULL, "%p", -1);
  LDParam(key, == NULL, "%p", -1);
  
  odata = hash_splay_remove(hash, key);
  
  if( odata != NULL ) {
    if( hash->free_func != NULL ) {
      hash->free_func( odata );
    }
    retval = 0;
  }

  return retval;
}

void* hash_splay_replace(hash_t* hash, void* key, void* data) {
  void* retval = NULL;
  
  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);
  LDParam(data, != data, "%p", NULL);

  retval = hash_splay_remove(hash, key);
  
  hash_splay_insert(hash, key, data);

  return retval;
}

void* hash_splay_lookup(hash_t* hash, void* key) {
  void* retval = NULL;
  s32_t index;
  
  LDFunc();
  LDParam(hash, == NULL, "%p", NULL);
  LDParam(key, == NULL, "%p", NULL);
    
  /* can we do this */
  index = hash->hash_func(key, hash->keylen, hash->buckets);

  if( index < 0 ) {
    errno = EINVAL;
  } else {
    retval = splay_lookup(hash->entries[index]->data, key);
  }

  return retval;
}

void** hash_splay_getkeys(hash_t* hash) {
  void** retval = NULL;
  
  LDFunc();
  
  return retval;
}

s32_t hash_splay_numentries(hash_t* hash) {
  s32_t retval = -1;
  
  LDFunc();
  LDParam(hash, == NULL, "%p", -1);

  return retval;
}

hash_ops_t splay_ops = {
  .initialize = hash_splay_initialize,
  .destroy = hash_splay_destroy,
  .insert = hash_splay_insert,
  .remove = hash_splay_remove,
  .delete = hash_splay_delete,
  .replace = hash_splay_replace,
  .lookup = hash_splay_lookup,
  .getkeys = hash_splay_getkeys,
  .numentries = hash_splay_numentries,
};

