/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LOUS_SRC_STRUCTURES_INTERNAL_HASH_H
#define LOUS_SRC_STRUCTURES_INTERNAL_HASH_H

#ident "@(#)hash.h 1.6      04/02/16 Chris G. Davis"

/** \file src/structures/internal/hash.h
 *  \brief Hash table internal definitions
 *  \author Chris G. Davis
 *
 */

/** \brief Hash entry structure
 *
 * An entry for each key/data pair in the hash
 */
struct LOUS_Hash_Entry {
  void* key; /**< key for this entry, allocated and freed by LOUS */
  void* data; /**< data for this entry, allocated and freed by free_func */
};


/** \brief Hash operations structure
 * 
 * This abstracts the operations for the different modes of hash operation
 *
 */
struct LOUS_Hash_Operations {
  /** initialize the hash entries 
   *
   * The hash must be predefined, but the entries should be NULL.
   *
   * \param hash the hash to initialize
   *
   * \return 0 if successful, -1 if failure, withe errno set
   */
  s32_t (*initialize)(hash_t* hash);

  /** destroy the allocated hash entries
   *
   * The hash must be predefined. Any keys are deleted, but data is not.
   *
   * \param hash the hash to destroy
   *
   * \return 0 if sucessful, -1 if failue, with errno set
   */
  s32_t (*destroy)(hash_t* hash);

  s32_t (*insert)(hash_t* hash, void* key, void* data);
  void* (*remove)(hash_t* hash, void* key);
  s32_t (*delete)(hash_t* hash, void* key);
  void* (*replace)(hash_t* hash, void* key, void* data);
  void* (*lookup)(hash_t* hash, void* key);
  void** (*getkeys)(hash_t*);
  s32_t (*numentries)(hash_t*);
};


/** \brief Hash table structure
 * 
 * Contains the hash table
 */
struct LOUS_Hash_Table {
  iterator_list_t iterate;

  /**
   * compare \a keya and \a keyb and returns the comparision
   */
  comp_t (*comp_func)(void* keya, void* keyb, s32_t keylen);
  s32_t keylen; /**< length of the keys, or 0 is the key is a string */

  s32_t buckets; /**< number of hash_buckets */
  s32_t num_entries; /**< number of entries in the hash */
  s32_t flags; /**< hash flags */

  hash_ops_t* ops; /**< which set of operations to use with this hash */

  hash_ent_t** entries; /**< the defined set of entries for this hash */

  /**
   * return the hash buck that \a key fits into
   */
  s32_t (*hash_func)(void* key, s32_t keylen, s32_t buckets);


  /**
   * delete data item
   */
  void (*free_func)(void* data);

};

/*
 *  the known operation classes
 */
extern hash_ops_t generic_ops;
extern hash_ops_t splay_ops;

#endif /* LOUS_SRC_STRUCTURES_INCLUDE_HASH_H */
