/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2001,2002,2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)iterator.c 1.1      04/02/16 Chris G. Davis"

/** \file src/structures/iterator.c
 *  \brief Hash table default routines
 *  \author Chris G. Davis
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lous/iterator.h"
#include "internal/iterator.h"

void iterator_do_remove(s32_t flags, iterator_list_t* list, iterator_entry_t* ent) {
  LDFunc();
  LDParam(list, == NULL, "%p");
  LDParam(ent, == NULL, "%p");

  if( list->first == ent ) {
    list->first = ent->next;
  }

  if( list->end == ent ) {
    list->end = ent->prev;
  }

  if( ent->prev != NULL ) {
    ent->prev->next = ent->next;
  }

  if( ent->next != NULL ) {
    ent->next->prev = ent->prev;
  }

}

void iterator_do_sorted_insert_ptr(iterator_list_comp_t* list, 
				   iterator_entry_key_ptr_t* ent, 
				   comp_t cdir) {
  iterator_entry_key_ptr_t* walker;

  LDFunc();
  LDParam(list, == NULL, "%p");
  LDParam(ent, == NULL, "%p");

  /* walk the list until we find an entry that returns cdir */
  if( list->first == NULL ) {
    list->first = (iterator_entry_t*)ent;
    list->end = (iterator_entry_t*)ent;
    return;
  }

  walker = (iterator_entry_key_ptr_t*)list->first;
  while( walker != NULL ) {
    if( list->comp_func(walker->key, ent->key, list->keylen) == cdir ) {
      /* we found one */
      ent->next = (iterator_entry_t*)walker;
      ent->prev = (iterator_entry_t*)walker->prev;
      if( walker->prev == NULL ) {
	list->first = (iterator_entry_t*)ent;
      } else {
	walker->prev->next = (iterator_entry_t*)ent;
      }

      walker->prev = (iterator_entry_t*)ent;

      return;
    }

    walker = (iterator_entry_key_ptr_t*)walker->next;
  }

  /* none found, add at end */
  ent->prev = list->end;
  list->end->next = (iterator_entry_t*)ent;
  ent->next = NULL;
  list->end = (iterator_entry_t*)ent;
}

void iterator_do_sorted_insert(iterator_list_comp_t* list, 
			       iterator_entry_key_t* ent, 
			       comp_t cdir) {
  iterator_entry_key_t* walker;

  LDFunc();
  LDParam(list, == NULL, "%p");
  LDParam(ent, == NULL, "%p");

  /* walk the list until we find an entry that returns cdir */
  if( list->first == NULL ) {
    list->first = (iterator_entry_t*)ent;
    list->end = (iterator_entry_t*)ent;
    return;
  }

  walker = (iterator_entry_key_t*)list->first;
  while( walker != NULL ) {
    if( list->comp_func(walker->key, ent->key, list->keylen) == cdir ) {
      /* we found one */
      ent->next = (iterator_entry_t*)walker;
      ent->prev = walker->prev;

      if( walker->prev == NULL ) {
	list->first = (iterator_entry_t*)ent;
      } else {
	walker->prev->next = (iterator_entry_t*)ent;
      }

      walker->prev = (iterator_entry_t*)ent;

      return;
    }

    walker = (iterator_entry_key_t*)walker->next;
  }

  /* none found, add at end */
  ent->prev = list->end;
  list->end->next = (iterator_entry_t*)ent;
  ent->next = NULL;
  list->end = (iterator_entry_t*)ent;

}

void iterator_do_insert(s32_t flags, iterator_list_t* list, iterator_entry_t* ent) {
  LDFunc();
  LDParam(list, == NULL, "%p");
  LDParam(ent, == NULL, "%p");

  switch( flags & LIST_ITERATOR ) {
  case LIST_KEEPFIFO:
    /* add to end */
    if(list->end != NULL ) {
      list->end->next = ent;
    } else {
      list->first = ent;
    }

    ent->prev = list->end;
    ent->next = NULL;

    list->end = ent;

    break;
  case LIST_KEEPLIFO:
    /* add to start */
    if(list->first != NULL ) {
      list->first->prev = ent;
    } else {
      list->end = ent;
    }

    ent->next = list->first;
    ent->prev = NULL;

    list->first = ent;

    break;
  case LIST_KEEPSORTED:
    /* sort */
    if( flags & LIST_KEYPTR ) {
      iterator_do_sorted_insert_ptr((iterator_list_comp_t*)list, (iterator_entry_key_ptr_t*)ent, greaterthen);
    } else {
      iterator_do_sorted_insert((iterator_list_comp_t*)list, (iterator_entry_key_t*)ent, greaterthen);
    }
    break;
  case LIST_KEEPREVERSE:
    /* reverse sort */
    if( flags & LIST_KEYPTR ) {
      iterator_do_sorted_insert_ptr((iterator_list_comp_t*)list, (iterator_entry_key_ptr_t*)ent, lessthen);
    } else {
      iterator_do_sorted_insert((iterator_list_comp_t*)list, (iterator_entry_key_t*)ent, lessthen);
    }
    break;
  case 0:
    break;
  }

}

iterator_t* iterator_do_alloc(iterator_list_t* list) {
  iterator_t* newit;

  LDFunc();
  LDParam(list, == NULL, "%p", NULL);

  Lmalloc(newit, sizeof( iterator_t ), NULL);

  newit->list = list;
  newit->prev = NULL;
  newit->last = NULL;
  newit->next = NULL;

  return newit;
}

void iterator_destroy(iterator_t* iter) {
  LDFunc();
  LDParam(iter, == NULL, "%p");

  free( iter );
}

void* iterator_getfirst(iterator_t* iter) {
  LDFunc();
  LDParam(iter, == NULL, "%p", NULL);

  return iter->list->first;

}

void* iterator_getlast(iterator_t* iter) {
  LDFunc();
  LDParam(iter, == NULL, "%p", NULL);

  return iter->list->end;
}

void* iterator_getnext(iterator_t* iter) {
  LDFunc();
  LDParam(iter, == NULL, "%p", NULL);

  /*
   * Cases:
   *  1 - we're a new iter (prev == next == last == NULL)
   *  2 - we last returned the first item ( prev == NULL, last == list->first, next == x)
   *  3 - we last returned the last item (prev == x, last == list->end; next == NULL )
   *
   * NOTE: in general, always ignore iter->last, since it may have been deleted
   */

  if( iter->last == NULL ) {
    /* new list */
    iter->last = iter->list->first;
    if( iter->last ) {
      iter->next = iter->last->next;
    }

    return iter->last;
  }

  if( iter->next != NULL ) {
    iter->prev = iter->next->prev;
    iter->last = iter->next;
    iter->next = iter->last->next;

    return iter->last;
  }

  return NULL;
}

void* iterator_getprev(iterator_t* iter) {
  LDFunc();
  LDParam(iter, == NULL, "%p", NULL);

  if( iter->last == NULL ) {
    /* new list */
    iter->last = iter->list->end;
    if( iter->last ) {
      iter->prev = iter->last->prev;
    }

    return iter->last;
  }

  if( iter->prev != NULL ) {
    iter->next = iter->prev->next;
    iter->last = iter->prev;
    iter->prev = iter->last->prev;

    return iter->last;
  }

  return NULL;
}

