/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)func.c 1.2      04/02/22 Chris G. Davis"

/** \file src/user/func.c
 *  \brief CLI function code
 *  \author Chris G. Davis 
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "lous/cli-macro.h"
#include "internal/cli.h"

cli_response_t* cli_response_alloc(s32_t max_lns, s32_t lw) {
  cli_response_t* retval = NULL;
  LDFunc();

  Lmalloc(retval, sizeof( cli_response_t*), NULL);

  if( ( retval->responses = queue_alloc(QUEUE_FIFO, NULL)) == NULL ) {
    LError("Unable to malloc response queue");
    free(retval);
    return NULL;
  }

  retval->response_lines = 0;
  retval->max_response_lines = max_lns;

  retval->max_line_width = lw;
  retval->line_width = 0;
  Lmalloc( retval->buffer, lw + 2, NULL, cli_response_destroy(retval); );

  retval->bptr = retval->buffer;

  retval->ret_code = 0;

  return retval;
}

void cli_response_destroy(cli_response_t* resp) {
  LDFunc();
  LDParam(resp, == NULL, "%p");

  cli_response_clear( resp );

  if( resp->buffer ) {
    free( resp->buffer );
  }

  if( resp->responses ) {
    queue_destroy( resp->responses );
  }

  free( resp );
}

void cli_response_clear(cli_response_t* resp ) {
  void* rdata;
  LDFunc();
  LDParam(resp, == NULL, "%p");

  memset(resp->buffer, 0, resp->max_line_width + 2);

  while( ( rdata = queue_pop( resp->responses )) != NULL ) {
    free( rdata );
  }

  resp->line_width = 0;
  resp->response_lines = 0;
  resp->bptr = resp->buffer;
}

void cli_response_set_retcode(cli_response_t* resp, s32_t code) {
  LDFunc();
  LDParam(resp, == NULL, "%p");

  resp->ret_code = code;
}

void cli_response_add_string(cli_response_t* resp, char* str, ...) {
  char lbuf[256];
  s32_t slen;
  va_list ap;

  LDFunc();
  LDParam(resp, == NULL, "%p");

  va_start(ap, str);
  vsnprintf(lbuf, 255, str, ap);
  va_end( ap );

  fflush(stdout);
  slen = strlen( lbuf ) + 1;

  if( slen > (resp->max_line_width - resp->line_width) ) {
    slen = (resp->max_line_width - resp->line_width);
  }

  snprintf(resp->bptr, slen, "%s", lbuf );
  resp->bptr += slen - 1;
  resp->line_width += slen - 1;
}

void cli_response_finish_string(cli_response_t* resp) {
  char* full;
  LDFunc();
  LDParam(resp, == NULL, "%p");

  Lstrdup(full, resp->buffer,);
  /* now take the string and push it */
  queue_push( resp->responses, full );
  resp->response_lines++;

  memset(resp->buffer, 0, resp->max_line_width + 2 );
  resp->bptr = resp->buffer;
}

char* cli_response_getline(cli_response_t* resp) {
  LDFunc();
  LDParam(resp, == NULL, "%p", NULL);

  return queue_pop( resp->responses );
}
