/*
 * Library of Useful Stuff (LOUS), a collection of useful functions
 * Copyright (C) 2004 Chris G. Davis <davis@lrtw.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ident "@(#)user.c 1.3      04/02/22 Chris G. Davis"

/** \file src/user/user.c
 *  \brief CLI user code
 *  \author Chris G. Davis 
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "lous/cli-macro.h"
#include "internal/cli.h"

cli_user_t* cli_user_alloc(cli_item_t* root, s32_t buflen, s32_t lines) {
  cli_user_t* retval = NULL;
  LDFunc();
  LDParam(root, == NULL, "%p", NULL );
  LDParam(buflen, <=0, "%d", NULL);

  Lmalloc(retval, sizeof( cli_user_t ), NULL );

  retval->max_line_width = buflen;
  retval->max_response_lines = lines;
  retval->tresp = NULL;
  retval->resp = NULL;
  retval->cli_root = root;

  return retval;
}

void cli_user_destroy(cli_user_t* user) {
  LDFunc();
  LDParam(user, == NULL, "%p");

  if( user->tresp ) {
    tokenizer_destroy_response(user->tresp );
  }

  if( user->resp ) {
    cli_response_destroy( user->resp );
  }

  free( user );
}

s32_t cli_user_run(cli_user_t* user, tokenizer_t* toker, char* string) {
  cli_item_t* cmd;
  s32_t error_level = 0;

  LDFunc();
  LDParam(user, == NULL, "%p", -1);
  LDParam(toker, == NULL, "%p", -1);
  /*
   * So, what we have is this:
   *  user has a char string that we want to run as a command
   *
   */

  /* ok, so lets tokenizer the buffer */
  if( user->tresp != NULL ) {
    tokenizer_destroy_response( user->tresp );
  }

  if( ( user->tresp = tokenizer_tokenize(toker, string)) == NULL ) {
    /* oops, bad shit */
    return -1;
  }
  
  if( user->resp == NULL ) {
    user->resp = cli_response_alloc(user->max_response_lines, user->max_line_width);
    if( user->resp == NULL ) {
      LError("Unable to alloc response");
      return -1;
    }
  } else {
    cli_response_clear( user->resp );
  }
  
  /* ok, so we have a command string, we have a tokenizer version of it */
  if( ( cmd = cli_match_token(user->cli_root, user->tresp, 0, &error_level)) == NULL ) {
    /* nothing matches, build an error response */
    cli_response_add_string(user->resp, "Error at token %d:", error_level + 1);
    cli_response_finish_string(user->resp);

    cli_response_add_string(user->resp, "  ");

    if( tokenizer_get_preceeding(user->tresp, error_level) != NULL ) {
      cli_response_add_string(user->resp, "%s ", tokenizer_get_preceeding(user->tresp, error_level));
    }

    cli_response_add_string(user->resp, "/%s\\", tokenizer_get_token(user->tresp, error_level) );

    if( tokenizer_get_remainder(user->tresp, error_level) != NULL ) {
      cli_response_add_string(user->resp, " %s", tokenizer_get_remainder(user->tresp, error_level));
    }      

    cli_response_finish_string(user->resp);
    cli_response_set_retcode( user->resp, CLI_COMMAND_NOTFOUND);

    return 0;
  }

  /* we know which command */
  cmd->command_func(user->resp, user->tresp);

  /* 
     ok, we tokenized, ran and got a response. now it's up to our caller
     to do something with it.
  */

  return 0;
}

cli_item_t* cli_match_token(cli_item_t* root, tokenizer_response_t* resp, s32_t level, s32_t* gotto) {
  cli_item_t* walker = NULL;
  iterator_t* iter;
  bool_t match;
  char* token;
  s32_t numtokens;

  LDFunc();
  LDParam(root, == NULL, "%p", NULL);
  LDParam(resp, == NULL, "%p", NULL);

  token = tokenizer_get_token(resp, level );
  numtokens = tokenizer_get_numtokens(resp);
  *gotto = level;

  /* ok, check for a match at this level */
  iter = iterator_alloc( root->items );
  while( ( walker = iterator_getnext(iter)) != NULL ) {
    match = false;

    if( walker->regex_string != NULL ) {
      /* match regex */
      if( regexec(&walker->regex, token, 0, NULL, 0) == 0 ) {
	match = true;
      }
    } else {
      if ( !strncasecmp(token, walker->tag, strlen( walker->tag )) ) {
	match = true;
      }
    }

    if( match == true ) {
      if( level + 1 == numtokens ) {
	/* the end */
	iterator_destroy( iter );
	return walker;
      }

      if( walker->items == NULL ) {
	/* nothing below this, but command has more data */
	iterator_destroy( iter );
	return walker;
      }

      iterator_destroy( iter );
      return cli_match_token(walker, resp, level + 1, gotto);
    }
  }
  iterator_destroy( iter );

  return NULL;
}

char* cli_user_response_getline(cli_user_t* user) {
  LDFunc();
  LDParam(user, == NULL, "%p", NULL);

  return cli_response_getline(user->resp);
}
