#
# Library of Useful Stuff (LOUS), a collection of useful functions
# Copyright (C) 2002,2004 Chris G. Davis <davis@lrtw.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#ident "@(#)test-header.sh 1.3      04/02/16 Chris G. Davis"

#####################################################
# Error and pass/fail handling
##################################################### 
showinfo () {
	echo "== stdout ===========================" >&4
	cat stdout.log >&4
	echo "== stderr ===========================" >&4
	cat stderr.log >&4
	echo "=====================================" >&4
}

showerror () {
	case "$LAST" in
	"compare")
		echo "== Difference $LAST1 $LAST2 ==========" >&4
		dodiff $LAST1.out.$LAST2 $LAST1.$LAST2 >&4
		;;
	*)
		showinfo
		;;
	esac
}

cleanup () {
    cd $OPWD
    if [ -z "$DONTCLEAN" ]; then
	rm -rf $TMPDIR
    fi
}

error () {
    echo "ERROR" >&4
    if [ ! -z "$*" ]; then
	echo $* >&5
    fi
    cleanup
    showinfo
    exit -1
}

pass () {
    echo "PASS" >&4
    if [ ! -z "$*" ]; then
	echo $* >&5
    fi
    cleanup
    exit 0
}

fail () {
    echo "FAIL" >&4
    if [ ! -z "$*" ]; then
	echo $* >&5
    fi
    showerror
    if [ -z "$DONTCLEANONFAIL" ]; then
        cleanup
    fi
    exit 1
}

#############################################################
# test commands
#############################################################
testfail () {
    if [ ! $? = 0 ]; then
	fail $@
    fi
}

testpass () {
    if [ $? = 0 ]; then
	fail $@
    fi
}

#############################################################
# compile subs
#############################################################
compile () {
	LAST="compile"
    if [ ! -f $1 ]; then
	fail "file $1 not found"
    fi

    gcc -I$INCLUDES $CFLAGS $TCFLAGS -c -o $2 $1
    testfail "compile $1 failed"
}

link () {
	LAST="link"
    out=$1
    shift
    gcc -L$LIBDIR $LDFLAGS -o $out $* $TLDFLAGS
    testfail "link $out failed"
}

run () {
    if [ ! -f $1 ]; then
	fail "file $1 does not exist"
    fi

    ./$1 >$1.out.stdout 2>$1.out.stderr
}

makerun () {
	out=$1
	shift
	SRCS=
	for fl in $*; do
		compile $fl.c $fl.o
		SRCS="$SRCS $fl.o"
	done

	link $out $SRCS

	runcomp $out
}

dodiff () {
    diff $1 $2
}

compare () {
    LAST="compare"
    LAST1=$1
    LAST2=$2
    if [ ! -f $1.$2 ]; then
	fail "expected output $1 not found"
    fi
    if [ ! -f $1.out.$2 ]; then
	fail "actual output $1 not found"
    fi

    dodiff $1.out.$2 $1.$2 >/dev/null 2>/dev/null
}

runcomp () {
    run $1 
    testfail "execution of $1 failed"

    compare $1 stdout
    testfail "$1 stdout wrong"

    compare $1 stderr
    testfail  "$1 stderr wrong"
}

saysection () {
	echo -n "$1 " >&4
}

#############################################################
# set test variables
#############################################################

if [ ! -d ../../scripts ]; then
    echo "Unable to find scripts dir"
    exit -1
fi

TREE=`../../scripts/findroot`
OPWD=$PWD
TESTCODE=`basename $PWD`
TMPROOT="/tmp"
TMPDIR=$TMPROOT/$TESTCODE-$$-$RANDOM

if [ -d $TMPDIR ]; then
    echo "Random test dir $TMPDIR already exists"
    exit -1
fi

############################################################
# get ready to run the test
############################################################

# ready tmpdir
mkdir -p $TMPDIR
cp * $TMPDIR 2>/dev/null
cd $TMPDIR

# redirect files
exec 4>&1
exec 5>&2
exec 1>stdout.log
exec 2>stderr.log

# compile flags
INCLUDES=$TREE/src/include
LIBDIR=$TREE/src/
CFLAGS="-g"
LDFLAGS=""

echo -n "[$TESTCODE] Test $TESTNAME: " >&4
